from TableauConMan.converters.converter import Converter
from TableauConMan.converters.project_converter import ProjectConverter
from TableauConMan.converters.permission_rule_converter import PermissionRuleConverter
from TableauConMan.sources.sources import Source
from loguru import logger


class ConverterFactory:
    @staticmethod
    def get_asset_converter(asset_type: str, asset_source: Source) -> Converter:
        if asset_type == "projects":
            asset_converter = ProjectConverter(asset_source)
        elif asset_type == "permission_rule":
            asset_converter = PermissionRuleConverter(asset_source)
        else:
            raise TypeError(f"No converter for the {asset_type} asset_type")

        return asset_converter



from TableauConMan.sources.sources import Source
from TableauConMan.assets.asset import Asset
from TableauConMan.managers.asset_manager import AssetManager
from TableauConMan.managers.permission_rule_manager import PermissionRuleManager
from TableauConMan.factories.converter_factory import ConverterFactory
import TableauConMan.config.plan_options as options

from typing import List, Tuple
from loguru import logger


class ProjectManager(AssetManager):
    def __init__(self) -> None:
        super().__init__(),
        self.permission_rule_manager = PermissionRuleManager()
        self.converter_factory = ConverterFactory()

    def add(
        self, asset_type: str, project_asset_list: List[Asset], target_source: Source
    ):
        """ """
        asset_converter = self.converter_factory.get_asset_converter(
            asset_type, target_source
        )

        for project_asset in project_asset_list:
            converted_project = asset_converter.convert_to_source(project_asset)
            created_project = target_source.add_item(asset_type, converted_project)

            logger.info(f"Project {created_project} added.")

            self.permission_rule_manager.add(
                target_source, created_project, project_asset
            )

    def remove(
        self, asset_type: str, project_item_list: List[Asset], target_source: Source
    ):
        if options.DELETE_ON_TARGET:
            project_item_list = self.get_source_list(
                asset_type, project_item_list, target_source
            )
            for project in project_item_list:
                target_source.delete_item(asset_type, project)

                logger.debug(f"Deleted project: {project}")

    def update(
        self,
        asset_type: str,
        project_item_list: List[Tuple[Asset, Asset]],
        target_source: Source,
    ):
        # logger.debug(f"Assets to Update: {project_item_list}")

        asset_converter = self.converter_factory.get_asset_converter(
            asset_type, target_source
        )

        for target_item, reference_item in project_item_list:
            target_item.update_asset(reference_item)

            converted_project = asset_converter.convert_to_source(target_item)

            updated_project = target_source.update_item(asset_type, converted_project)

            self.permission_rule_manager.update(
                target_source, updated_project, target_item, reference_item
            )

            logger.debug(f"Updated project: {updated_project}")



from TableauConMan.sources.sources import Source
from TableauConMan.assets.permission_rule_asset import PermissionRule
from TableauConMan.managers.asset_manager import AssetManager
from TableauConMan.factories.converter_factory import ConverterFactory


from loguru import logger


class PermissionRuleManager(AssetManager):
    def __init__(self) -> None:
        super().__init__()
        self.converter_factory = ConverterFactory()

    def add(
        self, target_source: Source, target_project_item, reference_project_asset
    ) -> None:
        asset_converter = self.converter_factory.get_asset_converter(
            "permission_rule", target_source
        )
        logger.debug(reference_project_asset)
        for permission_rule in reference_project_asset.permission_set:
            asset_type = permission_rule.asset_type

            converted_permission_rule = asset_converter.convert_to_source(
                permission_rule
            )

            target_source.add_sub_item(
                asset_type, target_project_item, converted_permission_rule
            )

            logger.info(f"Added {asset_type} Permissions for {target_project_item}.")

    def remove(self):
        logger.info(f"Removed Permissions")

    def update(
        self,
        target_source: Source,
        target_project_item,
        target_project_asset,
        reference_project_asset,
    ) -> None:
        asset_converter = self.converter_factory.get_asset_converter(
            "permission_rule", target_source
        )

        add, remove, update = self.get_update_lists(
            target_project_asset, reference_project_asset
        )

        for target_rule, reference_rule in update:
            in_a_and_b = sorted(
                list(
                    set(target_rule.capabilities.items())
                    & set(reference_rule.capabilities.items())
                )
            )
            in_a_not_b = sorted(
                list(set(target_rule.capabilities.items()) ^ set(in_a_and_b))
            )
            in_b_not_a = sorted(
                list(set(reference_rule.capabilities.items()) ^ set(in_a_and_b))
            )

            if len(in_a_not_b) > 0:
                remove_rule = PermissionRule(
                    asset_type=target_rule.asset_type,
                    grantee_name=target_rule.grantee_name,
                    grantee_type=target_rule.grantee_type,
                    capabilities=dict(),
                )

                for capability, mode in in_a_not_b:
                    remove_rule.capabilities.update({capability: mode})

                remove.append(remove_rule)

            if len(in_b_not_a) > 0:
                add_rule = PermissionRule(
                    asset_type=target_rule.asset_type,
                    grantee_name=target_rule.grantee_name,
                    grantee_type=target_rule.grantee_type,
                    capabilities=dict(),
                )

                for capability, mode in in_b_not_a:
                    add_rule.capabilities.update({capability: mode})

                add.append(add_rule)

        combined_update = add

        for permission_rule in combined_update:
            asset_type = permission_rule.asset_type
            converted_permission_rule = asset_converter.convert_to_source(
                permission_rule
            )
            logger.debug(converted_permission_rule)
            target_source.add_sub_item(
                asset_type, target_project_item, converted_permission_rule
            )

            logger.info(
                f"Updated {asset_type} permissions for {permission_rule.grantee_name} on {target_project_asset.project_name}."
            )

        for permission_rule in remove:
            asset_type = permission_rule.asset_type
            converted_permission_rule = asset_converter.convert_to_source(
                permission_rule
            )

            target_source.remove_sub_item(
                asset_type, target_project_item, converted_permission_rule
            )
            logger.info(
                f"Removed {asset_type} permissions for {permission_rule.grantee_name} on {target_project_asset.project_name}."
            )
        logger.info(f"Updated Permissions")

    def get_update_lists(self, target_project_item, reference_project_item):
        add, remove, update = self.group_asset_list(
            target_project_item.permission_set, reference_project_item.permission_set
        )
        """logger.info(
            f"Updated Permissions: Add: {add} Remove: {remove} Update: {update}"
        )"""

        return add, remove, update



from TableauConMan.plan import Plan
from TableauConMan.workbooks_manager import WorkbooksManager
from TableauConMan.data_sources_manager import DataSourcesManager
from TableauConMan.projects_manager import ProjectsManager
from TableauConMan.groups_manager import GroupsManager
from TableauConMan.yaml_connector import YamlConnector
from loguru import logger
from dotenv import load_dotenv
from . import cli
import click

load_dotenv()

# "./public_sync_plan.yaml"
@cli.command()  # type: ignore
@click.option(
    "--yaml_path",
    help="String path of where the yaml plan is located",
)
def migrate_content(yaml_path: str):
    """
    :param yaml_path: String path of where the yaml plan is located
    :return: None

    """
    raw_plan = YamlConnector(yaml_path)

    plan = Plan()
    plan.load_plan(raw_plan.get_yaml())
    logger.info("Loaded plan")

    migrate_datasources(plan)

    migrate_workbooks(plan)


@cli.command()  # type: ignore
@click.option(
    "--yaml_path",
    help="String path of where the yaml plan is located",
)
def provision_settings(yaml_path: str):
    """
    Provision Settings

    Provision settings based on a provided YAML plan.

    :param yaml_path: String path of where the YAML plan is located.
    :return: None
    """
    raw_plan = YamlConnector(yaml_path)

    plan = Plan()
    plan.load_plan(raw_plan.get_yaml())
    logger.info("Loaded plan")

    provision_groups(plan)


def migrate_datasources(migrate_plan: Plan):
    """
    :param migrate_plan: The migration plan containing the list of datasources to migrate.
    :type migrate_plan: Plan
    :return: None
    """
    logger.info("Processing Datasources")
    datasources = DataSourcesManager(migrate_plan)
    datasources.populate_datasources()
    logger.info("Populated datasources")
    # logger.debug(datasources.reference_datasources)

    datasource_options = datasources.get_datasource_options()

    to_update, to_add, to_remove = datasources.get_datasource_changes()
    logger.debug(f"Update:{to_update}, Add:{to_add}, Remove:{to_remove}")

    datasources.remove(to_remove)

    datasources.add(
        to_add,
        to_update,
        datasource_options,
    )

    logger.info("Processed Datasources")


def migrate_workbooks(migration_plan: Plan):
    """
    Migrates workbooks based on the given migration plan.

    :param migration_plan: The migration plan containing the information needed for workbook migration.
    :type migration_plan: Plan
    :return: None
    """
    logger.info("Processing Workbooks")
    workbooks = WorkbooksManager(migration_plan)
    workbooks.populate_workbooks()
    logger.info("Populated workbooks")
    logger.debug(workbooks.reference_workbooks)

    workbook_options = workbooks.get_workbook_options()
    to_update, to_add, to_remove = workbooks.get_workbook_changes()
    logger.debug(f"Update:{to_update}, Add:{to_add}, Remove:{to_remove}")

    workbooks.remove(to_remove)

    workbooks.add(to_add, to_update, workbook_options)

    logger.info("Processed Workbooks")


def provision_projects(provision_plan: Plan):
    """
    :param provision_plan: The provision plan object containing the necessary information for provisioning projects and permissions.
    :return: None

    This method takes a provision plan object and performs the following steps to provision projects:
    1. Creates an instance of ProjectsManager using the provision plan.
    2. Populates projects by calling populate_projects method of the ProjectsManager.
    3. Logs the populated projects, reference projects, target projects, target project paths and reference project paths.
    4. Retrieves project changes using get_project_changes method of ProjectsManager and logs the changes.
    5. Adds newly created projects using add method of ProjectsManager.
    6. Removes projects to be removed using remove method of ProjectsManager.
    7. Updates projects to be updated using update method of ProjectsManager.
    8. Gets project options using get_project_options method of ProjectsManager.
    9. If project options include 'update_permissions', the following additional steps are performed:
        a. Populates projects again.
        b. Populates users and groups using populate_users_and_groups method of ProjectsManager.
        c. Populates project permissions using populate_project_permissions method of ProjectsManager.
        d. Populates permission capabilities using populate_permission_capabilities method of ProjectsManager.
        e. Retrieves permission changes using get_permission_changes method of ProjectsManager and logs the changes.
        f. Adds new capabilities using add_capabilities method of ProjectsManager.
        g. Removes capabilities to be removed using remove_capabilities method of ProjectsManager.
    10. Logs the processed permissions.

    Note: The logger used in this method is assumed to be globally defined and available.
    """
    logger.info("Processing Projects")
    projects = ProjectsManager(provision_plan)
    projects.populate_projects()
    logger.info("Populated projects")
    logger.debug(f"Reference Projects: {projects.reference_projects}")
    logger.debug(f"Reference Project List: {projects.reference_projects_list}")
    logger.debug(f"Target Projects: {projects.target_projects}")
    logger.debug(f"Target Project List: {projects.target_projects_list}")
    logger.debug(f"Target Project Path List: {projects.target_project_paths_list}")
    logger.debug(
        f"Reference Project Path List: {projects.reference_project_paths_list}"
    )

    to_update, to_remove, to_add = projects.get_project_changes()
    logger.debug(f"Add:{to_add}, Remove:{to_remove}, Update: {to_update}")

    projects.add(to_add)

    projects.remove(to_remove)

    projects.update(to_update)

    logger.info("Processed Projects")

    project_options = projects.get_project_options()

    if project_options.get("update_permissions"):
        logger.info("Processing Permissions")
        projects.populate_projects()
        logger.info("Populated projects")
        projects.populate_users_and_groups()
        projects.populate_project_permissions()
        projects.populate_permission_capabilities()

        """logger.debug(
            f"Target Capabilities: {projects.target_project_capabilities_list}"
        )"""
        """logger.debug(
            f"Reference Capabilities: {projects.reference_project_capabilities_list}"
        )"""

        to_update, to_remove, to_add = projects.get_permission_changes()
        # logger.debug(f"Add:{to_add}, Remove:{to_remove}, Update: {to_update}")
        logger.debug(f"Add:{to_add}")
        logger.debug(f"Remove:{to_remove}")
        # logger.debug(f"Update: {to_update}")

        projects.add_capabilities(to_add)

        projects.remove_capabilities(to_remove)

    logger.info("Processed Permissions")


def provision_groups(provision_plan: Plan):
    """
    Provisions groups based on the given provision plan.

    :param provision_plan: The provision plan to use for provisioning groups.
    :return: None
    """
    logger.info("Processing Groups")
    groups = GroupsManager(provision_plan)
    groups.populate_groups()
    logger.info("Populated groups")
    logger.debug(f"Target Groups: {groups.target_groups_list}")
    logger.debug(f"Reference Groups: {groups.reference_groups_list}")
    logger.debug(f"Target Group Memberships: {groups.target_group_memberships_list}")
    logger.debug(
        f"Reference Group Memberships: {groups.reference_group_memberships_list}"
    )
    groups.populate_users()

    to_add, to_remove, to_update = groups.get_group_changes()
    logger.debug(f"Add:{to_add}, Remove:{to_remove}, Update: {to_update}")

    groups.remove(to_remove)

    groups.add(to_add)

    groups.update(to_update)

    logger.info("Processed Groups")



"""Classes and functions to convert between source and internal models of a project.
"""
from operator import itemgetter
from treelib import Tree
from typing import Union, List, Sized, Dict
import tableauserverclient as TSC
from loguru import logger

from TableauConMan.converters.converter import Converter
from TableauConMan.assets.project_asset import Project
from TableauConMan.sources.sources import Source
from TableauConMan.converters.permission_rule_converter import PermissionRuleConverter
from TableauConMan.assets.permission_rule_asset import PermissionRule
import TableauConMan.config.plan_options as options


class ProjectConverter(Converter):
    """Methods for converting between a TSC ProjectItem and TableauConMan Project asset"""

    def __init__(self, source: Source) -> None:
        """Including in the a Source on init to control item filtering

        :param source: Only using ServerSource, but kept at the abstract level
        """
        self.source: Source = source

    @property
    def project_path_id_map(self):
        """Maps the project_path property to the project_id property

        :return: A dict of project_path and project_id
        """
        return self._get_project_path_id_map()

    def convert_to_asset(self, source_item: Union[TSC.ProjectItem, dict]) -> Project:
        """Acts as a method factory to convert from the passed source

        :param item: Should be a project representation from the source.
        :raises TypeError: Will raise an error is there are no methods for the source_type
        :return: Returns a project item converted to the internal asset model.
        """
        converted_asset = None

        if self.source.source_type == "server":
            converted_asset = self.server_to_asset(source_item)
        elif self.source.source_type == "specification":
            converted_asset = self.specification_to_asset(source_item)
        else:
            raise TypeError(
                f"The source_type of {self.source.source_type} is not supported."
            )

        return converted_asset

    def convert_to_source(self, asset: Project) -> Union[TSC.ProjectItem, dict]:
        """Acts as a method factory to convert to the passed source

        :param item: Should be an internal project model
        :raises TypeError: Will raise an error is there are no methods for the source_type
        :return: Returns a source specific project item converted from the internal asset model.
        """
        converted_item = None
        if self.source.source_type == "server":
            converted_item = self.asset_to_server(asset)
        elif self.source.source_type == "specification":
            converted_item = self.asset_to_specification(asset)
        else:
            raise TypeError(
                f"The source_type of {self.source.source_type} is not supported."
            )

        return converted_item

    def asset_to_server(self, project_asset: Project) -> TSC.ProjectItem:
        """Converts a Project asset to a TSC.ProjectItem

        :param project_asset: The project asset to be converted
        :return: The TSC.ProjectItem representation of the project asset
        """
        server_project_item = TSC.ProjectItem(
            name=project_asset.project_name,
            description=project_asset.description,
            content_permissions=project_asset.content_permissions,
            parent_id=self._get_project_parent_id(
                project_asset.project_path, self.project_path_id_map
            ),
        )

        server_project_item._id = project_asset.project_id
        if options.INCLUDE_ASSET_PERMISSIONS and project_asset.manage_permissions:
            rule_converter = PermissionRuleConverter(self.source)

            for rule in project_asset.permission_set:
                if (
                    rule.asset_type
                    not in PermissionRule.PermissionAssetTypes._value2member_map_
                ):
                    continue
                rule_item = rule_converter.asset_to_server(rule)

                item_asset_rule_name = rule_converter._get_project_permission_name(
                    rule.asset_type
                )
                item_asset_rule_object = getattr(
                    server_project_item, item_asset_rule_name
                )
                if item_asset_rule_object is None:
                    setattr(
                        server_project_item, item_asset_rule_name, list([rule_item])
                    )
                else:
                    item_asset_rule_object.append(rule_item)

        return server_project_item

    def asset_to_specification(self, project_asset: Project) -> dict:
        """Converts a Project asset to a project Specification dict

        :param project_asset: The project asset to be converted
        :return: The Specification dict representation of the project asset
        """
        specification_project_item = dict(
            {
                "project_name": project_asset.project_name,
                "project_path": project_asset.project_path,
                "description": project_asset.description,
                "content_permissions": project_asset.content_permissions,
                "permission_set": [],
            }
        )

        if options.INCLUDE_ASSET_PERMISSIONS and project_asset.manage_permissions:
            rule_converter = PermissionRuleConverter(self.source)

            for rule in project_asset.permission_set:
                rule_item = rule_converter.asset_to_specification(rule)

                specification_permission_set: List[
                    Dict
                ] = specification_project_item.get("permission_set")
                if len(specification_permission_set) > 0:
                    set_rule: Dict = next(
                        item
                        for item in specification_permission_set
                        if (item.get("group_name") or item.get("user_name"))
                        == (rule_item.get("group_name") or rule_item.get("user_name"))
                    ).get("permission_rule")
                    set_rule.update(rule_item.get("permission_rule"))
                else:
                    specification_permission_set.append(rule_item)

        return specification_project_item

    def server_to_asset(self, project_item: TSC.ProjectItem) -> Project:
        """Converts a TSC.ProjectItem to a Project asset

        :param project_item: The TSC.ProjectItem to be converted
        :return: The Project asset representation of the TSC.ProjectItem
        """
        asset_project = Project(
            project_id=project_item.id,
            project_path=self._get_project_path(
                project_item.id, self.project_path_id_map
            ),
            project_name=project_item.name,
            description=project_item.description,
            content_permissions=project_item.content_permissions,
            parent_content_permissions=self._get_server_content_permissions(
                project_item.parent_id
            ),
            project_parent_id=project_item.parent_id,
        )

        if options.INCLUDE_ASSET_PERMISSIONS and asset_project.manage_permissions:
            rule_converter = PermissionRuleConverter(self.source)

            logger.debug(f"Getting Permissions for: {project_item.name}")

            for permission_asset in PermissionRule.PermissionAssetTypes:
                with self.source.connect():
                    populate_method = rule_converter._get_project_permission_method(
                        permission_asset.value, "populate"
                    )

                    populate_method(project_item)

                    permission_object = rule_converter._get_project_permission_object(
                        permission_asset.value, project_item
                    )

                    if permission_object is not None:
                        for rule in permission_object():
                            rule_item = rule_converter.server_to_asset(
                                rule, permission_asset.value
                            )

                            asset_project.permission_set.append(rule_item)
            asset_project.permission_set.sort(key=lambda item: item.specification_id)

        return asset_project

    def specification_to_asset(self, project_specification: dict) -> Project:
        """Converts a Specification dict to a Project asset

        :param project_specification: The Specification dict to be converted
        :return: The Project asset representation of the Specification dict
        """
        asset_project = Project(
            project_id="",
            project_name=project_specification.get("project_name"),
            project_path=project_specification.get("project_path"),
            description=project_specification.get("description"),
            content_permissions=project_specification.get("content_permissions"),
            project_parent_id="",
            parent_content_permissions=self._get_specification_parent_content_permissions(
                project_specification.get("project_path")
            ),
        )

        if options.INCLUDE_ASSET_PERMISSIONS and asset_project.manage_permissions:
            rule_converter = PermissionRuleConverter(self.source)

            logger.debug(f"Getting Permissions for: {asset_project.project_name}")

            permission_set: List[PermissionRule] = list()
            if not project_specification.get("permission_set"):
                asset_project.permission_set = permission_set

            for project_rule in project_specification.get("permission_set"):
                # logger.debug(f"Processing permission rule: {project_rule}")
                for asset_rule in project_rule.get("permission_rule").keys():
                    if (
                        asset_rule
                        not in PermissionRule.PermissionAssetTypes._value2member_map_
                    ):
                        continue
                    # logger.debug(f"Processing Rule: {asset_rule}")
                    permission_rule_asset = rule_converter.specification_to_asset(
                        project_rule, asset_rule
                    )

                    permission_set.append(permission_rule_asset)
                    # logger.debug(f"Added Rule: {permission_set}")

            asset_project.permission_set = permission_set

            asset_project.permission_set.sort(key=lambda item: item.specification_id)

        # logger.debug(f"Converted {project_specification} to {asset_project}")

        return asset_project

    @staticmethod
    def _get_project_path(project_item_id: str, project_path_id_map: dict) -> str:
        """Retrieves the project_path for a TSC.ProjectItem project_id

        :param project_item: The TSC.ProjectItem id to get the project_path for
        :param project_path_id_map: A dict of project_path and project_id
        :return: The project_path for the given TSC.ProjectItem
        """
        project_item_path = list(
            filter(
                lambda x: project_path_id_map[x] == project_item_id,
                project_path_id_map,
            )
        )[0]

        return project_item_path

    @staticmethod
    def _get_project_parent_id(project_path: str, project_path_id_map: dict):
        """Retrieves the project_id for project one level up from a given project_path

        :param project_path: A string representing the project_path of a project
        :param project_path_id_map: A dict of project_path and project_id
        :return: A project_id of the project one level up in the project_path
        """
        project_parent_path_end = project_path.rfind("/")
        # A top level project will not have a '/' in the path

        if project_parent_path_end == -1:
            project_parent_path = project_path
            parent_project_id = None
        else:
            project_parent_path = project_path[0:project_parent_path_end]
            parent_project_id = project_path_id_map.get(project_parent_path)

        return parent_project_id

    def _get_server_content_permissions(self, project_id: str) -> str:
        """Retrieves a mapping of project_id and content_permissions

        :return: A dict of project_id and content_permissions
        """
        project_item_list = self.source.get_assets("projects")

        item_content_permission = next(
            (
                item.content_permissions
                for item in project_item_list
                if item.id == project_id
            ),
            "",
        )

        return item_content_permission

    def _get_specification_parent_content_permissions(self, project_path: str) -> str:
        """Retrieves a mapping of project_path and content_permissions

        :return: A content_permissions
        """
        project_item_list = self.source.get_assets("projects")

        project_parent_path_end = project_path.rfind("/")

        if project_parent_path_end == -1:
            project_parent_path = ""
        else:
            project_parent_path = project_path[0:project_parent_path_end]

        item_content_permission = next(
            (
                item.get("content_permissions")
                for item in project_item_list
                if item.get("project_path") == project_parent_path
            ),
            "",
        )

        return item_content_permission

    def _get_project_path_id_map(self):
        """Retrieves a mapping of project_path and project_id

        :return: A dict of project_path and project_id
        """
        project_item_list = self.source.get_assets("projects")

        project_tree = self._parse_projects_to_tree(project_item_list)

        project_path_dict = {}

        for project in project_item_list:
            nodes = list(project_tree.rsearch(project.id))
            nodes.reverse()
            node_path = "/".join(
                [project_tree.get_node(node).data for node in nodes[1:]]
            )
            project_path_dict[node_path] = project.id

        return project_path_dict

    @staticmethod
    def _parse_projects_to_tree(project_item_list):
        """Constructs a node tree based on project_id and parent_id from TSC.ProjectItems

        :param project_item_list: A list of TSC.ProjectItems, should be a full server list for complete coverage
        :return: A tree object that can be searched for for project_id and project_name
        """
        tree = Tree()
        tree.create_node("tableau", "tableau", data="tableau")

        list_root = []
        for project in project_item_list:
            # logger.debug(project)
            if project.parent_id is None:  # not in project.keys()
                tree.create_node(
                    project.id, project.id, parent="tableau", data=project.name
                )
                list_root.append(project_item_list.index(project))

        # logger.debug(project_item_list)

        index_list = list(set(range(len(project_item_list))).difference(list_root))

        if len(index_list) > 0:
            project_list = list()
            project_subset = itemgetter(*index_list)(project_item_list)
            if len(index_list) == 1:
                project_list.append(project_subset)
            else:
                project_list.extend(project_subset)

            while len(index_list) > 0:
                remove_index = list()
                for index, curr_project in enumerate(project_list):
                    parent_project_id = curr_project.parent_id
                    child_project_id = curr_project.id
                    child_name = curr_project.name
                    if tree.get_node(parent_project_id) is not None:
                        tree.create_node(
                            child_project_id,
                            child_project_id,
                            parent=parent_project_id,
                            data=child_name,
                        )
                        remove_index.append(index)

                index_list = list(
                    set(range(len(project_list))).difference(remove_index)
                )
                if len(index_list) > 0:
                    project_list = list(itemgetter(*index_list)(project_list))
        return tree



"""A module to set up the connections to a Tableau server"""
from typing import Optional
import tableauserverclient as TSC
from loguru import logger


class ServerConnector:
    """
    This class provides a connector to a server using Tableau Server Client (TSC) library.

    Args:
        site_name (str): The name of the site on the server.
        server_url (str): The URL of the server.
        token_name (str): The name of the personal access token.
        token_secret (str): The secret of the personal access token.

    Attributes:
        server_name (Optional[str]): The name of the server.
        server_url (str): The URL of the server.
        token_name (str): The name of the personal access token.
        token_secret (str): The secret of the personal access token.
        authentication (Optional[TSC.PersonalAccessTokenAuth]): The authentication object.
        server (Optional[TSC.Server]): The server object.

    Methods:
        connect(): Connects to the server using the provided credentials.

    Properties:
        server (Optional[TSC.Server]): The server object property.
        authentication (Optional[TSC.PersonalAccessTokenAuth]): The authentication object property.
    """

    def __init__(self, site_name, server_url, token_name, token_secret):
        """

        :param site_name:
        :param server_url:
        :param token_name:
        :param token_secret:
        """
        self.server_name: Optional[str] = site_name.strip()
        self.server_url: str = server_url.strip()
        self.token_name: str = token_name.strip()
        self.token_secret: str = token_secret.strip()
        self.authentication: Optional[TSC.PersonalAccessTokenAuth] = None
        self.server: Optional[TSC.Server] = None

    def __repr__(self):
        """

        :return:
        """
        return "<Server {server_name} server_object={_server} auth_object={_authentication}>".format(
            **self.__dict__
        )

    @property
    def server(self) -> Optional[TSC.Server]:
        """

        :return:
        """
        return self._server

    @server.setter
    def server(self, value):
        """

        :param value:
        """
        if value is None:
            self._server = TSC.Server(self.server_url, use_server_version=True)
        elif not isinstance(value, TSC.Server):
            raise TypeError("server must be of TSC.Server class")
        else:
            self._server = value

    @property
    def authentication(self) -> Optional[TSC.PersonalAccessTokenAuth]:
        """

        :return:
        """
        return self._authentication

    @authentication.setter
    def authentication(self, value):
        """

        :param value:
        """
        if value is None:
            self._authentication = TSC.PersonalAccessTokenAuth(
                self.token_name,
                self.token_secret,
                self.server_name,
            )
        elif not isinstance(value, TSC.PersonalAccessTokenAuth):
            raise TypeError(
                "authentication must be of  TSC.PersonalAccessTokenAuth class"
            )
        else:
            self._authentication = value

    def connect(self):
        """

        :return:
        """
        connection = self.server.auth.sign_in(self.authentication)
        return connection



from dataclasses import dataclass
from abc import ABC, abstractproperty


@dataclass
class Asset(ABC):
    @abstractproperty
    def specification_id(self):
        pass

    @abstractproperty
    def update_checksum(self):
        pass

    @abstractproperty
    def update_value(self):
        pass



import os
import hashlib
import json
from typing import Optional
from TableauConMan.assets_manager import AssetsManager
from TableauConMan.helpers import utils
from operator import itemgetter
from treelib import Tree
import tableauserverclient as TSC
from loguru import logger


class ProjectsManager(AssetsManager):
    """
    Initialize the ProjectsManager object.

    :param plan: The plan object specifying the target and reference servers.
    """

    class PermissionAssets:
        Workbook = "workbook"
        Project = "project"
        Datasource = "datasource"
        Flow = "flow"
        Lens = "lens"
        Datarole = "datarole"
        Metric = "metric"

    VALID_PERMISSION_ASSETS = [
        PermissionAssets.Datasource,
        PermissionAssets.Workbook,
        PermissionAssets.Project,
        PermissionAssets.Flow,
        PermissionAssets.Lens,
        PermissionAssets.Datarole,
        PermissionAssets.Metric,
    ]

    def __init__(self, plan) -> None:
        """

        :param plan:
        """
        AssetsManager.__init__(self, plan)
        self.target_projects: Optional[list[TSC.ProjectItem]] = None
        self.target_projects_list: Optional[list] = None
        self.target_project_paths_list: Optional[list] = None
        self.reference_projects: Optional[list[TSC.ProjectItem]] = None
        self.reference_projects_list: Optional[list] = None
        self.reference_project_paths_list: Optional[list] = None

    def populate_projects(self):
        """
        Use object properties to get a comparable list of workbooks from the source and reference
        """
        project_options = self.get_project_options()
        raw_target_projects, raw_target_projects_list = self.generate_server_list(
            self.plan.target,
            self.AssetType.Projects,
        )

        raw_target_project_paths_list = self.get_project_paths(raw_target_projects)

        spec = self.plan.reference

        spec.projects

        # logger.debug(spec.projects)

        reference_projects_list = list(x.get("project_name") for x in spec.projects)
        reference_project_paths_list = list(
            x.get("project_path") for x in spec.projects
        )

        if project_options.get("manage_only_spec_projects"):
            target_project_paths_list_filtered = {
                key: value
                for key, value in raw_target_project_paths_list.items()
                if key in reference_project_paths_list
            }  # {k:v for (k,v) in d.items() if filter_string in k}

            logger.debug(
                f"Filters target Project paths: {target_project_paths_list_filtered}"
            )

            target_projects_filtered = utils.get_filtered_item_list(
                raw_target_projects, target_project_paths_list_filtered.values(), "id"
            )

            logger.debug(f"Filtered target Projects: {target_projects_filtered}")

            target_projects_list_filtered = [x.name for x in target_projects_filtered]

            logger.debug(
                f"Filtered target Project list: {target_projects_list_filtered}"
            )

            target_projects = target_projects_filtered
            target_projects_list = target_projects_list_filtered
            target_project_paths_list = target_project_paths_list_filtered

        else:
            target_projects = raw_target_projects
            target_projects_list = raw_target_projects_list
            target_project_paths_list = raw_target_project_paths_list

        self.target_projects = target_projects
        self.target_projects_list = target_projects_list
        self.target_project_paths_list = target_project_paths_list

        self.reference_projects_list = reference_projects_list
        self.reference_project_paths_list = reference_project_paths_list

    def populate_users_and_groups(self):
        with self.plan.target.connect():
            self.target_users = list(TSC.Pager(self.plan.target.server.users))
            self.target_groups = list(TSC.Pager(self.plan.target.server.groups))

    def populate_project_permissions(self):
        project_server = self.plan.target.server.projects
        with self.plan.target.connect():
            for project in self.target_projects:
                for asset_type in self.VALID_PERMISSION_ASSETS:
                    permission_method = self.get_project_permission_method(
                        asset_type, project_server, "populate"
                    )
                    permission_method(project)

                logger.info(f"Project permissions for {project} have been populated")

    def populate_permission_capabilities(self):
        (
            server_capabilities_list,
            server_capabilities,
        ) = self.get_server_capabilities_list()

        spec_capabilities_list, spec_capabilities = self.get_spec_capabilities_list()

        self.target_project_capabilities = server_capabilities
        self.target_project_capabilities_list = server_capabilities_list
        self.reference_project_capabilities = spec_capabilities
        self.reference_project_capabilities_list = spec_capabilities_list

        # logger.debug(f"Target Capabilities: {server_capabilities}")
        # logger.debug(f"Reference Capabilities: {spec_capabilities}")

    def generate_server_project_schema(self):
        raw_target_projects, raw_target_projects_list = self.generate_server_list(
            self.plan.target,
            self.AssetType.Projects,
        )

        raw_target_project_paths_list = self.get_project_paths(raw_target_projects)

        projects = dict({"projects": []})

        permission_template_list = list()

        for project in raw_target_projects:
            project_path = self.get_project_item_path(
                project, raw_target_project_paths_list
            )

            permission_set, capabilities = self.get_project_item_permission_set(project)

            permission_template_list.extend(capabilities)

            project_schema = dict(
                {
                    "project_name": project.name,
                    "project_path": project_path,
                    "project_id": project.id,
                    "description": project.description,
                    "content_permissions": project.content_permissions,
                    "permission_set": permission_set,
                }
            )

            projects.get("projects").append(project_schema)

        # logger.debug(f"Permission Template List: {permission_template_list}")

        unique_permission_templates = list(
            {
                v.get("permission_template_id"): v for v in permission_template_list
            }.values()
        )

        # logger.debug(f"Unique Permission Templates: {unique_permission_templates}")

        permission_templates = dict(
            {"permission_templates": unique_permission_templates}
        )

        return projects, permission_templates

    def get_project_item_path(self, project_item, project_path_map):
        project_path = list(
            filter(
                lambda x: project_path_map[x] == project_item.id,
                project_path_map,
            )
        )[0]
        return project_path

    def generate_server_permission_templates_schema(self, permission_template_list):
        pass

    def get_project_item_permission_set(self, project_item):
        self.populate_users_and_groups()

        server_users = self.target_users
        server_groups = self.target_groups

        permission_set = list()  # dict({"permission_set": []})

        with self.plan.target.connect():
            for asset_type in self.VALID_PERMISSION_ASSETS:
                permission_method = self.get_project_permission_method(
                    asset_type, self.plan.target.server.projects, "populate"
                )
                permission_method(project_item)
                permission_object = self.get_project_permission_object(
                    asset_type, project_item
                )
                # permission_list = permission_set.get("permission_set")

                # logger.debug(f"Permission Object: {permission_object}")
                for permission in permission_object:
                    # permission_rule = dict()
                    if permission.grantee.tag_name == "user":
                        user_item = list(
                            filter(
                                lambda x: x.id == permission.grantee.id,
                                server_users,
                            )
                        )[0]

                        grantee_name = user_item.name
                        grantee_key = "user_name"

                    elif permission.grantee.tag_name == "group":
                        group_item = list(
                            filter(
                                lambda x: x.id == permission.grantee.id,
                                server_groups,
                            )
                        )[0]
                        grantee_name = group_item.name
                        grantee_key = "group_name"

                    # permission_rule.update({asset_type: permission.capabilities})

                    permission_rule = dict(
                        {
                            grantee_key: grantee_name,
                            "permission_list": {asset_type: permission.capabilities},
                        }
                    )

                    existing_grantee = next(
                        (
                            item
                            for item in permission_set
                            if item.get(grantee_key) == grantee_name
                        ),
                        None,
                    )

                    if existing_grantee:
                        existing_grantee.get("permission_list").update(
                            {asset_type: permission.capabilities}
                        )
                    else:
                        permission_set.append(permission_rule)

        capabilities = list()

        for rule in permission_set:
            rule_list = rule.get("permission_list").copy()
            rule_id = hashlib.md5(
                json.dumps(rule_list, sort_keys=True).encode("utf-8")
            ).hexdigest()
            rule.update({"permission_template_id": rule_id})

            rule_list.update({"permission_template_id": rule_id})

            capabilities.append(rule_list)

            rule.pop("permission_list")

        """logger.debug(
            f"Project: {project_item.name} |  Permission Rules: {permission_set}"
        )"""

        return permission_set, capabilities

    def get_inherited_grantees(self, project):
        inherited_grantees = []
        for permission in project.permissions:
            # logger.debug(f"Permission Capabilities:{permission.capabilities} ")
            if permission.capabilities.get("InheritedProjectLeader") is not None:
                if permission.capabilities.get("InheritedProjectLeader") == "Allow":
                    inherited_grantees.append(permission.grantee.id)

        return inherited_grantees

    def get_server_capabilities_list(self):
        logger.info("Processing Capabilities")
        server = self.plan.target.server
        server_projects = self.target_projects
        server_users = self.target_users
        server_groups = self.target_groups
        project_options = self.get_project_options()  # Filter based on options

        server_project_path_map = self.target_project_paths_list

        server_capabilities_list = list()
        server_capabilities_dict_list = list()

        for project in server_projects:
            project_path = list(
                filter(
                    lambda x: server_project_path_map[x] == project.id,
                    server_project_path_map,
                )
            )[0]
            # project_path_dict = dict({project_path: list()})
            # server_capabilities_list.append(project_path_dict)
            """logger.debug(
                f"Capabilities Project: {project} | Path: {project_path} | capabilities list: {server_capabilities_list} "
            )"""
            """logger.warning(
                f"Project {project.name} | Content Permissions: {project.content_permissions} | Project Parent ID: {project.parent_id}"
            )"""
            if project.parent_id:
                parent_project_content_permissions = list(
                    filter(lambda x: project.parent_id == x.id, server_projects)
                )[0].content_permissions
            else:
                parent_project_content_permissions = "None"
            """logger.debug(
                f"Project Content Permission:{project.content_permissions} | Parent Project Content Permissions: {parent_project_content_permissions} "
            )"""
            if (
                project.content_permissions != project.ContentPermissions.ManagedByOwner
                and (
                    project.content_permissions
                    == project.ContentPermissions.LockedToProject
                    and parent_project_content_permissions
                    != project.ContentPermissions.LockedToProject
                )
                or project.content_permissions
                == project.ContentPermissions.LockedToProjectWithoutNested
            ):
                # Check for inheritance and filer out permissions
                with self.plan.target.connect():
                    inherited_grantees = self.get_inherited_grantees(project)
                    for asset_type in self.VALID_PERMISSION_ASSETS:
                        permission_object = self.get_project_permission_object(
                            asset_type, project
                        )
                        for permission in permission_object:
                            # print(project.id,permission.grantee.tag_name,permission.capabilities)
                            if permission.grantee.id not in inherited_grantees:
                                if permission.grantee.tag_name == "user":
                                    user_item = list(
                                        filter(
                                            lambda x: x.id == permission.grantee.id,
                                            server_users,
                                        )
                                    )[0]

                                    grantee_name = user_item.name
                                elif permission.grantee.tag_name == "group":
                                    group_item = list(
                                        filter(
                                            lambda x: x.id == permission.grantee.id,
                                            server_groups,
                                        )
                                    )[0]
                                    grantee_name = group_item.name

                                for capability in permission.capabilities.keys():
                                    server_capability_detail = {}

                                    server_capability_detail.update(
                                        {
                                            "project_path": project_path,
                                            "grantee_type": permission.grantee.tag_name,
                                            "grantee_name": grantee_name,
                                            "asset_type": asset_type,
                                            "capability": capability,
                                            "mode": permission.capabilities.get(
                                                capability
                                            ),
                                        }
                                    )

                                    server_capabilities_dict_list.append(
                                        server_capability_detail
                                    )

                                    server_capabilities_list.append(
                                        f"{project_path} | {permission.grantee.tag_name} | {grantee_name} | {asset_type} | {capability} | {permission.capabilities.get(capability)}"
                                    )

        return server_capabilities_list, server_capabilities_dict_list

    def get_spec_capabilities_list(self):
        spec_projects = self.plan.reference.projects
        # spec_projects = list(filter(lambda x: x['project_path'] == 'Pilot/Business Insights', spec_projects))

        spec_permission_templates = self.plan.reference.permission_templates

        # print(spec_permission_templates)

        spec_capabilities_list = list()
        spec_capabilities_dict_list = list()

        for project in spec_projects:
            project_content_permissions = project.get("content_permissions")
            permission_set = project.get("permission_set")
            project_path = project.get("project_path")
            parent_project_path = project_path.split(project.get("project_name"))[
                0
            ].rstrip("/")
            if parent_project_path != "":
                parent_project = list(
                    filter(
                        lambda x: parent_project_path == x.get("project_path"),
                        spec_projects,
                    )
                )[0]
            else:
                parent_project = dict()

            """logger.debug(
                f"Project Parent Path: {parent_project_path} | Project Path: {project_path} | Parent Project: {parent_project}"
            )"""
            parent_project_content_permissions = parent_project.get(
                "content_permissions"
            )
            if (
                permission_set is not None
                and project_content_permissions
                != TSC.ProjectItem.ContentPermissions.ManagedByOwner
                and (
                    project_content_permissions
                    == TSC.ProjectItem.ContentPermissions.LockedToProject
                    and parent_project_content_permissions
                    != TSC.ProjectItem.ContentPermissions.LockedToProject
                )
                or project_content_permissions
                == TSC.ProjectItem.ContentPermissions.LockedToProjectWithoutNested
            ):
                # print(project['permission_set'])
                for permission in permission_set:
                    if permission.get("group_name") is not None:
                        grantee_type = "group"
                        grantee_name = permission.get("group_name")
                    elif permission.get("user_name") is not None:
                        grantee_type = "user"
                        grantee_name = permission.get("user_name")
                    # print(permission['permission_rule'])
                    try:
                        permission_template = list(
                            filter(
                                lambda x: x.get("name")
                                == permission.get("permission_rule"),
                                spec_permission_templates,
                            )
                        )[0]
                    except Exception as exc:
                        logger.info(
                            f"Project {project.get('project_path')} does not have a valid permission template: {permission.get('permission_rule')} "
                        )
                        logger.debug(exc)
                    # print(permission_template)
                    for asset_type, capabilities in permission_template.items():
                        # print(asset_type,capabilities, permission['permission_rule'])
                        if (
                            asset_type != "name"
                            and asset_type in self.VALID_PERMISSION_ASSETS
                        ):
                            # print(asset_type,capabilities)
                            for capability, mode in capabilities.items():
                                # print(capability,asset_type[capability])
                                spec_capability_detail = {}

                                spec_capability_detail.update(
                                    {
                                        "project_path": project.get("project_path"),
                                        "grantee_type": grantee_type,
                                        "grantee_name": grantee_name,
                                        "asset_type": asset_type,
                                        "capability": capability,
                                        "mode": mode,
                                    }
                                )

                                # print(server_capability_detail)
                                spec_capabilities_dict_list.append(
                                    spec_capability_detail
                                )
                                spec_capabilities_list.append(
                                    f"{project.get('project_path')} | {grantee_type} | {grantee_name} | {asset_type} | {capability} | {mode}"
                                )

        return spec_capabilities_list, spec_capabilities_dict_list

    def get_project_permission_object(self, asset_type, project):
        if asset_type == self.PermissionAssets.Project:
            permission_object_name = "permissions"
        else:
            permission_object_name = f"default_{asset_type}_permissions"

        permission_object = getattr(project, permission_object_name)

        return permission_object

    def get_project_permission_method(self, asset_type, server_projects, action):
        if asset_type == self.PermissionAssets.Project:
            if action == "update":
                permission_method_name = "update_permission"
            elif action == "delete":
                permission_method_name = "delete_permission"
            elif action == "populate":
                permission_method_name = "populate_permissions"
        else:
            if action == "update":
                permission_method_name = f"update_{asset_type}_default_permissions"
            elif action == "delete":
                permission_method_name = f"delete_{asset_type}_default_permissions"
            elif action == "populate":
                permission_method_name = f"populate_{asset_type}_default_permissions"

        permission_method = getattr(server_projects, permission_method_name)

        return permission_method

    def get_project_changes(self):
        """ """

        common, to_remove, to_add = self.get_changes(
            self.target_project_paths_list, self.reference_project_paths_list
        )

        to_update = list()

        for project in common:
            (
                target_project_content_permission,
                target_project_description,
                reference_project_content_permission,
                reference_project_description,
                target_project_item,
            ) = self.get_project_update_attributes(project)

            if (
                target_project_content_permission
                != reference_project_content_permission
                or target_project_description != reference_project_description
            ):
                to_update.append(project)

        return to_update, to_remove, to_add

    def get_project_update_attributes(self, project_path):
        target_project_id = self.target_project_paths_list.get(project_path)
        target_project_item = utils.get_filtered_item_list(
            self.target_projects, target_project_id, "id"
        )[0]
        target_project_content_permission = target_project_item.content_permissions
        target_project_description = target_project_item.description

        reference_project = utils.get_filtered_dict_list(
            self.plan.reference.projects, project_path, "project_path"
        )[0]

        reference_project_content_permission = reference_project.get(
            "content_permissions"
        )
        reference_project_description = reference_project.get("description")

        return (
            target_project_content_permission,
            target_project_description,
            reference_project_content_permission,
            reference_project_description,
            target_project_item,
        )

    def get_permission_changes(self):
        """ """
        common, to_remove, to_add = self.get_changes(
            self.target_project_capabilities_list,
            self.reference_project_capabilities_list,
        )

        return common, to_remove, to_add

    def get_project_paths(self, project_item_list):
        project_tree = self.parse_projects_to_tree(project_item_list)

        project_path_dict = {}

        for project in project_item_list:
            nodes = list(project_tree.rsearch(project.id))
            nodes.reverse()
            node_path = "/".join(
                [project_tree.get_node(node).data for node in nodes[1:]]
            )
            project_path_dict[node_path] = project.id
            # project_path_dict[project.id] = f"{node_path}"

            # project_json["projects"].append(project_path_dict)

        return project_path_dict

    def parse_projects_to_tree(self, project_item_list):
        tree = Tree()
        tree.create_node("tableau", "tableau", data="tableau")

        list_root = []
        for project in project_item_list:
            if project.parent_id is None:  # not in project.keys()
                tree.create_node(
                    project.id, project.id, parent="tableau", data=project.name
                )
                list_root.append(project_item_list.index(project))

        index_list = list(set(range(len(project_item_list))).difference(list_root))
        # print(f"The projects object {index_list}" )
        """logger.debug(
            f"The index list {index_list} | The project_item_list: {project_item_list} | The list_root: {list_root}"
        )"""
        # logger.debug(f"The project list: {itemgetter(*index_list)(project_item_list)}")
        if len(index_list) > 0:
            project_list = list()
            project_subset = itemgetter(*index_list)(project_item_list)
            if len(index_list) == 1:
                project_list.append(project_subset)
            else:
                project_list.extend(project_subset)

            while len(index_list) > 0:
                remove_index = list()
                for index, curr_project in enumerate(project_list):
                    # print(f"The projects object {curr_project}" )
                    parent_project_id = curr_project.parent_id  # ['@parentProjectId']
                    child_project_id = curr_project.id  # ['@id']
                    child_name = curr_project.name  # ['@name']
                    if tree.get_node(parent_project_id) is not None:
                        tree.create_node(
                            child_project_id,
                            child_project_id,
                            parent=parent_project_id,
                            data=child_name,
                        )
                        remove_index.append(index)

                index_list = list(
                    set(range(len(project_list))).difference(remove_index)
                )
                if len(index_list) > 0:
                    project_list = list(
                        itemgetter(*index_list)(project_list)
                    )  # list(project_list.itemgetter(*index_list)) #
        return tree

    def get_project_parent_id(self, project_path):
        project_parent_path_end = project_path.rfind("/")
        # A top level project will not have a '/' in the path

        """logger.debug(
            f"Project path end: {project_parent_path_end} | Project path {project_path} | target paths: {self.target_project_paths_list}"
        )"""

        if project_parent_path_end == -1:
            project_parent_path = project_path
            parent_project_id = None
        else:
            project_parent_path = project_path[0:project_parent_path_end]
            parent_project_id = self.target_project_paths_list.get(project_parent_path)

        return parent_project_id

    def format_reference_project(self, project_path):
        reference_project = dict()

        parent_project_id = self.get_project_parent_id(project_path)

        project_details = utils.get_filtered_dict_list(
            self.plan.reference.projects, project_path, "project_path"
        )[0]

        project_details.update({"parent_project_id": parent_project_id})

        reference_project = project_details

        # logger.debug(reference_project)

        return reference_project

    def create(self, reference_project):
        new_project = TSC.ProjectItem(
            name=reference_project.get("project_name"),
            content_permissions=reference_project.get("content_permissions"),
            description=reference_project.get("description"),
            parent_id=reference_project.get("parent_project_id"),
        )

        with self.plan.target.connect():
            created_project = self.plan.target.server.projects.create(new_project)

        return created_project

    def get_project_options(self):
        """

        :return:
        """
        project_config = self.plan.asset_options.get("projects")
        project_options = project_config.get("options")
        return project_options

    def update_project(self, project_item):
        with self.plan.target.connect():
            updated_project = self.plan.target.server.projects.update(project_item)

        return updated_project

    def delete(self, project_item):
        with self.plan.target.connect():
            self.plan.target.server.projects.delete(project_item.id)

    def update_permission(self, project_item, asset_type, rules_list):
        source = self.plan.target
        asset_permissions = self.get_project_permission_method(
            asset_type, source.server.projects, "update"
        )

        with source.connect():
            asset_permissions(project_item, rules_list)

        logger.info(f"Project {project_item.name} permissions updated")

    def delete_permission(self, project_item, asset_type, rules_list):
        source = self.plan.target
        asset_permissions = self.get_project_permission_method(
            asset_type, source.server.projects, "delete"
        )

        # logger.debug(f"Project: {project_item} | Rules: {rules_list}")

        with source.connect():
            # try:
            asset_permissions(project_item, rules_list[0])
        # except Exception as exc:
        #     logger.warning(exc)

        logger.info(f"Project {project_item.name} permissions deleted")

    def add(self, to_add_list):
        """ """

        for project in to_add_list:
            reference_project = self.format_reference_project(project)

            created_project = self.create(reference_project)

            logger.info(f"Project {created_project} added.")

    def remove(self, to_remove_list):
        project_options = self.get_project_options()

        # logger.debug(f"Target Project Paths: {self.target_project_paths_list}")

        if not project_options.get("manage_only_spec_projects"):
            for project in to_remove_list:
                project_id = self.target_project_paths_list.get(project)
                project_item = list(
                    filter(lambda x: project_id == x.id, self.target_projects)
                )[0]
                if project_options.get("archive_on_delete"):
                    archive_project = project_options.get("archive_project_path")
                    archive_project_id = self.target_project_paths_list.get(
                        archive_project
                    )
                    # Check for if the project is already in the Archived location
                    if project_item.parent_id != archive_project_id:
                        project_item.parent_id = archive_project_id

                        archived_project = self.update_project(project_item)

                        logger.debug(f"Archived project: {archived_project}")
                else:
                    self.delete(project_item)

                    logger.debug(f"Deleted project: {project_item}")

    def update(self, to_update_list):
        for project in to_update_list:
            (
                target_project_content_permission,
                target_project_description,
                reference_project_content_permission,
                reference_project_description,
                project_item,
            ) = self.get_project_update_attributes(project)

            project_item.content_permissions = reference_project_content_permission
            project_item.description = reference_project_description

            updated_project = self.update_project(project_item)

            logger.debug(f"Updated project: {updated_project}")

    def format_project_permission_rules(
        self, project_path, asset_type, capabilities_list
    ):
        project = project_path

        asset_capabilities_list = utils.get_filtered_dict_list(
            capabilities_list, asset_type, "asset_type"
        )

        logger.debug(
            f"Project {project} Asset: {asset_type} capabilities list: {asset_capabilities_list}"
        )

        grantee_list = list(
            set([x.get("grantee_name") for x in asset_capabilities_list])
        )

        project_asset_grantee_rules = list()

        for grantee in grantee_list:
            grantee_asset_capabilities_list = utils.get_filtered_dict_list(
                asset_capabilities_list, grantee, "grantee_name"
            )

            logger.debug(
                f"Project {project} Asset: {asset_type} Grantee: {grantee} capabilities list: {grantee_asset_capabilities_list}"
            )

            capabilities = dict()
            for capability in grantee_asset_capabilities_list:
                capabilities.update(
                    {
                        getattr(
                            TSC.Permission.Capability,
                            capability.get("capability"),
                        ): getattr(TSC.Permission.Mode, capability.get("mode"))
                    }
                )

            grantee_type = capability.get("grantee_type")
            if grantee_type == "group":
                grantee_item_list = self.target_groups
            elif grantee_type == "user":
                grantee_item_list = self.target_users
            grantee_item = self.get_item_from_list(grantee, grantee_item_list)

            project_asset_grantee_rule = TSC.PermissionsRule(
                grantee=grantee_item, capabilities=capabilities
            )

            project_asset_grantee_rules.append(project_asset_grantee_rule)

        logger.debug(
            f"Project Asset Grantee Rules: {project}-{asset_type}-{project_asset_grantee_rules}"
        )
        return project_asset_grantee_rules

    def format_capability(self, capability_string):
        capability_list = capability_string.split(" | ")
        capability_dict = dict()
        capability_project_path = capability_list[0]
        capability_grantee_type = capability_list[1]
        capability_grantee_name = capability_list[2]
        capability_asset_type = capability_list[3]
        capability = capability_list[4]
        mode = capability_list[5]
        capability_dict.update(
            {
                "project_path": capability_project_path,
                "grantee_type": capability_grantee_type,
                "grantee_name": capability_grantee_name,
                "asset_type": capability_asset_type,
                "capability": capability,
                "mode": mode,
            }
        )

        return capability_dict

    def add_capabilities(self, to_add_list):
        """
        Filter list by project
        Get project item
        Get group items
        get user items
        format rules
        """
        for capability in to_add_list:
            formatted_capability = self.format_capability(capability)
            capability_list = list()
            capability_list.append(formatted_capability)

            logger.debug(f"Formatted capability {formatted_capability}")

            project = formatted_capability.get("project_path")

            project_id = self.target_project_paths_list.get(project)
            project_item = list(
                filter(lambda x: project_id == x.id, self.target_projects)
            )[0]
            asset_type = formatted_capability.get("asset_type")

            project_rules = self.format_project_permission_rules(
                project, asset_type, capability_list
            )

            self.update_permission(project_item, asset_type, project_rules)

    def remove_capabilities(self, to_remove_list):
        for capability in to_remove_list:
            formatted_capability = self.format_capability(capability)
            capability_list = list()
            capability_list.append(formatted_capability)

            logger.debug(f"Formatted capability {formatted_capability}")

            project = formatted_capability.get("project_path")

            project_id = self.target_project_paths_list.get(project)
            project_item = list(
                filter(lambda x: project_id == x.id, self.target_projects)
            )[0]
            asset_type = formatted_capability.get("asset_type")

            project_rules = self.format_project_permission_rules(
                project, asset_type, capability_list
            )

            self.delete_permission(project_item, asset_type, project_rules)



"""Classes and functions to convert between source and internal models of a project permission rule.
"""
from typing import List
import tableauserverclient as TSC


from TableauConMan.converters.converter import Converter
from TableauConMan.assets.permission_rule_asset import PermissionRule
from TableauConMan.sources.sources import Source
from TableauConMan.sources.server_source import ServerSource


class PermissionRuleConverter(Converter):
    """Methods for converting between a TSC ProjectItem and ConMan ProjectSpec"""

    def __init__(self, source: Source) -> None:
        """Including in the a Source on init to control item filtering

        :param source: Only using ServerSource, but kept at the abstract level
        """
        self.source: Source = source

    def convert_to_asset(
        self, source_item: TSC.PermissionsRule or dict, asset_type: str
    ) -> PermissionRule:
        """_summary_

        :param source_item: _description_
        :return: _description_
        """
        converted_asset = None

        if isinstance(source_item, TSC.PermissionsRule):
            converted_asset = self.server_to_asset(source_item, asset_type)
        else:
            converted_asset = self.specification_to_asset(source_item, asset_type)

        return converted_asset

    def convert_to_source(self, asset: PermissionRule) -> TSC.PermissionsRule or dict:
        """_summary_

        :param asset: _description_
        :return: _description_
        """
        converted_item = None

        if isinstance(self.source, ServerSource):
            converted_item = self.asset_to_server(asset)
        else:
            converted_item = self.asset_to_specification(asset)

        return converted_item

    def asset_to_server(
        self, permission_rule_asset: PermissionRule
    ) -> TSC.PermissionsRule:
        """
        convert grantee name to TSC id

        """
        server_permission_rule = TSC.PermissionsRule(
            grantee=self._get_grantee_item(
                permission_rule_asset.grantee_name, permission_rule_asset.grantee_type
            ),
            capabilities=permission_rule_asset.capabilities,
        )

        return server_permission_rule

    def asset_to_specification(self, permission_rule_asset: PermissionRule) -> dict:
        """_summary_

        :param permission_rule_asset: _description_
        :return: _description_
        """
        specification_permission_rule = dict(
            {
                permission_rule_asset.grantee_type
                + "_name": permission_rule_asset.grantee_name,
                "permission_rule": dict(
                    {
                        permission_rule_asset.asset_type: permission_rule_asset.capabilities
                    }
                ),
            }
        )

        return specification_permission_rule

    def server_to_asset(
        self, permission_rule_item: TSC.PermissionsRule, asset_type: str
    ) -> PermissionRule:
        """_summary_

        :param permission_rule_item: _description_
        :param asset_type: _description_
        :return: _description_
        """
        permission_rule_asset = PermissionRule(
            asset_type=asset_type,
            grantee_name=self._get_grantee_name(
                permission_rule_item.grantee.id, permission_rule_item.grantee.tag_name
            ),
            grantee_type=permission_rule_item.grantee.tag_name,
            capabilities=permission_rule_item.capabilities,
        )

        return permission_rule_asset

    def specification_to_asset(
        self, specification_permission_rule: dict, asset_type: str
    ) -> PermissionRule:
        """_summary_

        :param specification_permission_rule: _description_
        :param asset_type: _description_
        :return: _description_
        """
        permission_rule_asset = PermissionRule(
            asset_type=asset_type,
            grantee_name=specification_permission_rule.get("group_name")
            or specification_permission_rule.get("user_name"),
            grantee_type=next(
                item[: len(item) - 5]
                for item in specification_permission_rule.keys()
                if item[-5:] == "_name"
            ),
            capabilities=specification_permission_rule.get("permission_rule", {}).get(
                asset_type
            ),
        )

        return permission_rule_asset

    def _get_grantee_item(self, grantee_name: str, grantee_type: str):
        """summary

        :param grantee_name: _description_
        :param grantee_type: _description_
        :raises TypeError: _description_
        :return: _description_
        """

        if grantee_type == "user":
            grantee_item_list = self.source.get_assets("users")
        elif grantee_type == "group":
            grantee_item_list = self.source.get_assets("groups")
        else:
            raise TypeError

        grantee_item = next(
            item for item in grantee_item_list if item.name == grantee_name
        )

        return grantee_item

    def _get_grantee_name(self, grantee_id: str, grantee_type: str) -> str:
        """summary

        :param grantee_name: _description_
        :param grantee_type: _description_
        :raises TypeError: _description_
        :return: _description_
        """

        if grantee_type == "user":
            grantee_item_list = self.source.get_assets("users")
        elif grantee_type == "group":
            grantee_item_list = self.source.get_assets("groups")
        else:
            raise TypeError

        grantee_name = next(
            item.name for item in grantee_item_list if item.id == grantee_id
        )

        return grantee_name

    def _convert_capabilities(self, asset_capabilities: List[dict]) -> dict:
        """_summary_

        :param asset_capabilities: _description_
        :return: _description_
        """
        capabilities_dict = dict(
            {item.capability: item.mode for item in asset_capabilities}
        )

        return capabilities_dict

    def _get_project_permission_object(self, asset_type, project_item):
        """_summary_

        :param asset_type: _description_
        :param project_item: _description_
        :return: _description_
        """
        if asset_type == "project":
            permission_object_name = "_permissions"
        else:
            permission_object_name = f"_default_{asset_type}_permissions"

        permission_object = getattr(project_item, permission_object_name)

        return permission_object

    def _get_project_permission_name(self, asset_type):
        """_summary_

        :param asset_type: _description_
        :return: _description_
        """
        if asset_type == "project":
            permission_object_name = "_permissions"
        else:
            permission_object_name = f"_default_{asset_type}_permissions"

        return permission_object_name

    def _get_project_permission_method(self, asset_type, action):
        """_summary_

        :param asset_type: _description_
        :param action: _description_
        :return: _description_
        """
        if asset_type == "project":
            if action == "update":
                permission_method_name = "update_permission"
            elif action == "delete":
                permission_method_name = "delete_permission"
            elif action == "populate":
                permission_method_name = "populate_permissions"
        else:
            if action == "update":
                permission_method_name = f"update_{asset_type}_default_permissions"
            elif action == "delete":
                permission_method_name = f"delete_{asset_type}_default_permissions"
            elif action == "populate":
                permission_method_name = f"populate_{asset_type}_default_permissions"

        permission_method = getattr(self.source.server.projects, permission_method_name)

        return permission_method



from TableauConMan.plan import Plan
from TableauConMan.workbooks_manager import WorkbooksManager
from TableauConMan.data_sources_manager import DataSourcesManager
from TableauConMan.projects_manager import ProjectsManager
from TableauConMan.groups_manager import GroupsManager
from TableauConMan.user_manager import UserManager
from TableauConMan.yaml_connector import YamlConnector

from TableauConMan.config.plans import Plan as plan_v2
from TableauConMan.factories.manager_factory import AssetManagerFactory
import TableauConMan.config.plan_options as options

from loguru import logger
from dotenv import load_dotenv
from . import cli
import click

load_dotenv()


@cli.command("provision-users", help="Provision users")  # type: ignore
@click.option(
    "--yaml_path",
    help="String path of where the yaml plan is located",
)
def provision_users(yaml_path: str):
    """

    :param yaml_path:
    :return:
    """
    raw_plan = YamlConnector(yaml_path)

    plan = Plan()
    plan.load_plan(raw_plan.get_yaml())
    logger.info("Loaded plan")
    process_user_provisioning(plan)


# "./public_sync_plan.yaml"
@cli.command("migrate-content", help="Migrate content from one site to another.")  # type: ignore
@click.option(
    "--yaml_path",
    help="String path of where the yaml plan is located",
)
def migrate_content(yaml_path: str):
    """

    :param yaml_path:
    :return:
    """
    raw_plan = YamlConnector(yaml_path)

    plan = Plan()
    plan.load_plan(raw_plan.get_yaml())
    logger.info("Loaded plan")

    migrate_datasources(plan)

    migrate_workbooks(plan)


@cli.command("provision-settings", help="Provision settings")  # type: ignore
@click.option(
    "--yaml_path",
    help="String path of where the yaml plan is located",
)
def provision_settings(yaml_path: str):
    """

    :param yaml_path:
    :return:
    """
    raw_plan = YamlConnector(yaml_path)

    plan = Plan()
    plan.load_plan(raw_plan.get_yaml())
    logger.info("Loaded plan")

    provision_groups(plan)


@cli.command()
@click.option(
    "--plan-file",
    "-p",
    type=click.Path(),
    help="String path of where the yaml plan is located",
)
def configure_assets(plan_file: str):
    """_summary_

    :param plan_file: _description_
    """
    raw_plan = YamlConnector(plan_file)

    file_text_list = raw_plan.get_yaml()

    # Load

    plan = plan_v2().load(file_text_list)
    logger.info("Loaded plan")

    load_project_options(plan.project_options)

    configure_projects(plan)


def migrate_datasources(migrate_plan: Plan):
    datasources = DataSourcesManager(migrate_plan)
    logger.info("Processing Datasources")
    datasources.populate_datasources()
    logger.info("Populated datasources")
    # logger.debug(datasources.reference_datasources)

    datasource_options = datasources.get_datasource_options()

    to_update, to_add, to_remove = datasources.get_datasource_changes()
    logger.debug(f"Update:{to_update}, Add:{to_add}, Remove:{to_remove}")

    datasources.remove(to_remove)

    datasources.add(
        to_add,
        to_update,
        datasource_options,
    )

    logger.info("Processed Datasources")


def migrate_workbooks(migration_plan: Plan):
    logger.info("Processing Workbooks")
    workbooks = WorkbooksManager(migration_plan)
    workbooks.populate_workbooks()
    logger.info("Populated workbooks")
    logger.debug(workbooks.reference_workbooks)

    workbook_options = workbooks.get_workbook_options()
    to_update, to_add, to_remove = workbooks.get_workbook_changes()
    logger.debug(f"Update:{to_update}, Add:{to_add}, Remove:{to_remove}")

    workbooks.remove(to_remove)

    workbooks.add(to_add, to_update, workbook_options)

    logger.info("Processed Workbooks")


def provision_projects(provision_plan: Plan):
    logger.info("Processing Projects")
    projects = ProjectsManager(provision_plan)
    projects.populate_projects()
    logger.info("Populated projects")
    logger.debug(f"Reference Projects: {projects.reference_projects}")
    logger.debug(f"Reference Project List: {projects.reference_projects_list}")
    logger.debug(f"Target Projects: {projects.target_projects}")
    logger.debug(f"Target Project List: {projects.target_projects_list}")
    logger.debug(f"Target Project Path List: {projects.target_project_paths_list}")
    logger.debug(
        f"Reference Project Path List: {projects.reference_project_paths_list}"
    )

    to_update, to_remove, to_add = projects.get_project_changes()
    logger.debug(f"Add:{to_add}, Remove:{to_remove}, Update: {to_update}")

    projects.add(to_add)

    projects.remove(to_remove)

    projects.update(to_update)

    logger.info("Processed Projects")

    project_options = projects.get_project_options()

    if project_options.get("update_permissions"):
        logger.info("Processing Permissions")
        projects.populate_projects()
        logger.info("Populated projects")
        projects.populate_users_and_groups()
        projects.populate_project_permissions()
        projects.populate_permission_capabilities()

        """logger.debug(
            f"Target Capabilities: {projects.target_project_capabilities_list}"
        )"""
        """logger.debug(
            f"Reference Capabilities: {projects.reference_project_capabilities_list}"
        )"""

        to_update, to_remove, to_add = projects.get_permission_changes()
        # logger.debug(f"Add:{to_add}, Remove:{to_remove}, Update: {to_update}")
        logger.debug(f"Add:{to_add}")
        logger.debug(f"Remove:{to_remove}")
        # logger.debug(f"Update: {to_update}")

        projects.add_capabilities(to_add)

        projects.remove_capabilities(to_remove)

    logger.info("Processed Permissions")


def configure_projects(configure_plan: plan_v2):
    for asset_type in options.ConfigureAssets:
        if asset_type.value in configure_plan.assets:
            asset_manager = AssetManagerFactory.get_asset_manager(asset_type.value)

            target_asset_list = asset_manager.get_asset_list(
                asset_type.value, configure_plan.target
            )

            reference_asset_list = asset_manager.get_asset_list(
                asset_type.value, configure_plan.reference
            )

            logger.info(f"Loaded Target and Reference for {asset_type}.")

            (
                assets_to_add,
                assets_to_remove,
                assets_to_update,
            ) = asset_manager.group_asset_list(target_asset_list, reference_asset_list)

            asset_manager.add(asset_type.value, assets_to_add, configure_plan.target)

            asset_manager.remove(
                asset_type.value, assets_to_remove, configure_plan.target
            )

            asset_manager.update(
                asset_type.value, assets_to_update, configure_plan.target
            )

        logger.info(f"Completed processing of {asset_type.value}")


def provision_groups(provision_plan: Plan):
    logger.info("Processing Groups")
    groups = GroupsManager(provision_plan)
    groups.populate_groups()
    logger.info("Populated groups")
    logger.debug(f"Target Groups: {groups.target_groups_list}")
    logger.debug(f"Reference Groups: {groups.reference_groups_list}")
    logger.debug(f"Target Group Memberships: {groups.target_group_memberships_list}")
    logger.debug(
        f"Reference Group Memberships: {groups.reference_group_memberships_list}"
    )
    groups.populate_users()

    to_add, to_remove, to_update = groups.get_group_changes()
    logger.debug(f"Add:{to_add}, Remove:{to_remove}, Update: {to_update}")

    groups.remove(to_remove)

    groups.add(to_add)

    groups.update(to_update)

    logger.info("Processed Groups")


def process_user_provisioning(provision_plan: Plan):
    """

    :param provision_plan:
    :return:
    """
    user_manager = UserManager(provision_plan)
    server_user_dict = user_manager.populate_server_users()

    reference_users = provision_plan.reference.users
    reference_user_dict = {
        f"{r.get('user_name')}@{r.get('user_name_domain')}": r.get("site_role")
        for r in reference_users
    }

    to_add, to_remove, to_update = user_manager.compare_dicts(
        reference_user_dict, server_user_dict
    )

    logger.debug(f"Add:{to_add}, Remove:{to_remove}, Update: {to_update}")

    [user_manager.create_user(a, reference_user_dict.get(a)) for a in to_add]
    [user_manager.update_user(a[0], a[1]) for a in to_update]
    [user_manager.delete_user(a) for a in to_remove]

    logger.info("Processed users")


def load_project_options(project_options: dict) -> None:
    options.INCLUDE_ASSET_PERMISSIONS = project_options.get("include_asset_permissions")



from dataclasses import dataclass
from typing import List
from enum import Enum
from TableauConMan.assets.asset import Asset
import hashlib
import json


@dataclass
class PermissionRule(Asset):
    class PermissionAssetTypes(Enum):
        Project = "project"
        Workbook = "workbook"
        Datasource = "datasource"

    asset_type: str
    grantee_type: str
    grantee_name: str
    capabilities: dict

    @property
    def specification_id(self):
        return hashlib.md5(
            (self.asset_type + self.grantee_type + self.grantee_name).encode("utf-8")
        ).hexdigest()

    @property
    def update_checksum(self):
        return hashlib.md5(
            (
                self.specification_id + json.dumps(self.capabilities, sort_keys=True)
            ).encode("utf-8")
        ).hexdigest()

    @property
    def update_value(self):
        return 1



import os
from typing import Optional
from TableauConMan.assets_manager import AssetsManager
from TableauConMan.helpers import utils
import tableauserverclient as TSC
from loguru import logger


class DataSourcesManager(AssetsManager):
    """
    DataSourcesManager class is responsible for managing data sources in the target and reference servers.

    To use the DataSourcesManager class, first create an instance by passing in a plan object. The plan object should contain the necessary information regarding the target and reference
    * servers, such as server connection and selection rules.

    Example usage:

    plan = {
        'target': target_server,
        'reference': reference_server,
        'target_selection_rules': target_selection_rules,
        'reference_selection_rules': reference_selection_rules,
        'asset_options': {
            'datasources': {
                'mapping': datasource_mapping,
                'custom_options': datasource_custom_options
            }
        }
    }

    data_sources_manager = DataSourcesManager(plan)

    After creating an instance of DataSourcesManager, you can use the following methods:

    1. populate_datasources()
       - This method populates the target and reference data sources lists based on the selection rules provided in the plan object.

    2. get_all_datasources(server_type: str)
       - This method retrieves data sources from a specified server type (target or reference).

    3. get_workbook_datasources(workbook_items)
       - This method retrieves workbook datasources based on the provided workbook items.

    4. get_datasource_options()
       - This method gets the data source options from the plan object.

    5. get_datasource_changes()
       - This method compares the target and reference data source lists and returns the changes that need to be made (update, add, remove).

    6. delete(server, datasource)
       - This method deletes a data source from the specified server.

    7. download(server, datasource)
       - This method downloads a data source from the specified server.

    8. format_target_datasource(reference_datasource)
       - This method formats a reference data source to match the target data source format.

    9. get_target_project()
       - This method gets the target project based on the plan asset project_type option.

    10. upload(server, datasource, file_path)
        - This method uploads a data source to the specified server.

    11. add(list_to_add: list, list_to_update: list = [], datasource_options: dict = {})
        - This method adds or updates data sources from the provided list.

    Please refer to the inline comments in the code for more details on each method and its parameters.
    """

    def __init__(self, plan) -> None:
        """

        :param plan:
        """
        AssetsManager.__init__(self, plan)
        self.target_datasources: Optional[list[TSC.DatasourceItem]] = None
        self.target_datasources_list: Optional[list] = None
        self.reference_datasources: Optional[list[TSC.DatasourceItem]] = None
        self.reference_datasources_list: Optional[list] = None

    def populate_datasources(self):
        """
        Use object properties to get a comparable list of workbooks from the source and reference.
        """

        (
            self.target_datasources,
            self.target_datasources_list,
        ) = self._generate_server_list(
            self.plan.target,
            self.AssetType.Datasources,
            self.plan.target_selection_rules,
        )

        (
            self.reference_datasources,
            self.reference_datasources_list,
        ) = self.get_all_datasources("reference")

    def get_all_datasources(self, server_type: str):
        """
        Can retrieve multiple data sources with the same name.
        Have to collect the data sources from the other Items that can reference them: Workbooks and Flows
        """
        source_server = getattr(self.plan, server_type)
        source_selection_rules = getattr(self.plan, server_type + "_selection_rules")
        datasources, datasource_list = self._generate_server_list(
            source_server,
            self.AssetType.Datasources,
            source_selection_rules,
        )

        workbooks, workbooks_list = self._generate_server_list(
            source_server,
            self.AssetType.Workbooks,
            source_selection_rules,
        )

        logger.debug(f"Workbook Name List: {workbooks_list}")

        (
            workbook_datasource_ids,
            workbook_datasource_names,
        ) = self.get_workbook_datasources(workbooks)

        all_datasources, all_datasources_list = self._generate_server_list(
            source_server,
            self.AssetType.Datasources,
        )

        workbook_datasources = utils.get_filtered_item_list(
            all_datasources, workbook_datasource_ids, "id"
        )

        with source_server.connect():
            for datasource in workbook_datasources:
                logger.debug(
                    f"Datasource: {datasource.name.replace(' ','')} has {datasource.content_url} content_url"
                )

        """
        Add process to check against the mapping or server request options
        """

        logger.debug(f"Workbook Datasource Names: {workbook_datasource_names}")

        combined_datasources = set(datasources + workbook_datasources)

        combined_datasources_list = set(datasource_list + workbook_datasource_names)

        return combined_datasources, combined_datasources_list

    def get_workbook_datasources(self, workbook_items):
        workbook_ids = list()

        for workbook in workbook_items:
            workbook_ids.append(workbook.id)

        logger.debug(f"List of Workbook ids: {workbook_ids}")

        workbook_datasource_ids = list()

        workbook_datasource_names = list()

        try:
            with self.plan.reference.connect():
                # Execute the query
                query_result = self.plan.reference.server.metadata.query(
                    """
                query findWorkbookDatasource($luids: [String]) {
                  workbooksConnection(filter: {luidWithin: $luids}) {
                    nodes{
                      name
                      upstreamDatasources {
                        name
                        luid
                      }
                    }
                  }
                }
                """,
                    {"luids": workbook_ids},
                )

                logger.debug(f"Metadata query result: {query_result}")

                result_data = query_result.get("data")
                data_nodes = result_data.get("workbooksConnection")
                workbook_list = data_nodes.get("nodes")

                for workbook in workbook_list:
                    for datasource in workbook.get("upstreamDatasources"):
                        logger.debug(
                            f"Workbook: {workbook} | Datasource: {datasource.get('name')} | luid: {datasource.get('luid')} "
                        )
                        workbook_datasource_ids.append(datasource.get("luid"))
                        workbook_datasource_names.append(datasource.get("name"))

        except Exception as exc:
            logger.warning(
                f"There was a problem processing GraphQL for workbooks: {workbook_ids}: {exc}"
            )

        return workbook_datasource_ids, workbook_datasource_names

    def _generate_server_list(
        self, source, asset_type, request_filter: Optional[TSC.RequestOptions] = None
    ):
        """

        :param source:
        :param asset_type:
        :param filter:
        :return:
        """
        asset = getattr(source.server, asset_type)

        with source.connect():
            asset_items = list(TSC.Pager(asset, request_filter))
            """
          Generalized: can be moved to parent class
          Find a way to included updated_at check.  Might be able to use dict to
          """

            asset_list = []
            for asset in asset_items:
                asset_list.append(asset.name)
            return asset_items, asset_list

    def get_datasource_options(self):
        """

        :return:
        """
        datasource_options = self.plan.asset_options.get("datasources")
        return datasource_options

    def get_datasource_changes(self):
        """

        :return:
        """
        common, to_remove, to_add = self.get_changes(
            self.target_datasources_list, self.reference_datasources_list
        )

        update = []
        for datasource_name in common:
            target_datasource = self.get_item_from_list(
                datasource_name, self.target_datasources
            )

            reference_datasource = self.get_item_from_list(
                datasource_name, self.reference_datasources
            )

            logger.debug(
                f"Datasource {reference_datasource.name} | Reference Updated At: {reference_datasource.updated_at.strftime('%Y-%m-%d, %H:%M:%S')} | Target Updated At {target_datasource.updated_at.strftime('%Y-%m-%d, %H:%M:%S')}"
            )

            if reference_datasource.updated_at > target_datasource.updated_at:
                update.append(datasource_name)

        return update, to_add, to_remove

    def delete(self, server, datasource):
        """

        :param server:
        :param datasource:
        """
        with server.connect():
            server.server.datasources.delete(datasource.id)

    def download(self, server, datasource):
        """

        :param server:
        :param datasource:
        :return:
        """
        with server.connect():
            file_path = server.server.datasources.download(datasource.id)

        return file_path

    def format_target_datasource(self, reference_datasource):
        """

        :param reference_datasource:
        :return:
        """
        target_datasource = TSC.DatasourceItem("")
        target_datasource.name = reference_datasource.name
        # target_datasource.content_url = reference_datasource.content_url
        target_datasource.project_id = self.get_target_project()

        return target_datasource

    def get_target_project(self):
        """
        Get the target project based on the plan asset project_type option
        """
        datasource_options = self.get_datasource_options()
        datasource_mapping = datasource_options.get("mapping")
        request_options = self.plan.format_rule(
            datasource_mapping.get("project_filter")
        )
        server = self.plan.target
        with server.connect():
            server_project = self.plan.target.server.projects.get(request_options)[0][0]
        target_project = server_project.id
        return target_project

    def upload(self, server, datasource, file_path):
        """

        :param server:
        :param datasource:
        :param file_path:
        :return:
        """
        with server.connect():
            new_workbook = server.server.datasources.publish(
                datasource, file_path, "Overwrite"
            )

        logger.debug(
            f"Datasource {new_workbook.name} has a content url of: {new_workbook.content_url}"
        )

        return new_workbook

    def add(
        self,
        list_to_add: list,
        list_to_update: list = [],
        datasource_options: dict = {},
    ):
        """
        Match the workbook from list to reference object
        Download reference workbook
        Get source project object
        Set workbook project to source project_id
        Upload workbook
        Delete downloaded workbook file
        """
        to_add_and_update = list_to_add + list_to_update

        if len(to_add_and_update) > 0:
            for datasource_name in to_add_and_update:
                logger.info(f"Processing Addition of Datasource: {datasource_name}")
                reference_datasource = self.get_item_from_list(
                    datasource_name, self.reference_datasources
                )

                file_path = self.download(self.plan.reference, reference_datasource)
                target_datasource = self.format_target_datasource(reference_datasource)
                uploaded_datasource = self.upload(
                    self.plan.target, target_datasource, file_path
                )

                logger.debug(
                    f"Reference content URL: {reference_datasource.content_url} | Uploaded content URL: {uploaded_datasource.content_url}"
                )
                if reference_datasource.content_url != uploaded_datasource.content_url:
                    logger.warning(
                        f"The reference content URL,{reference_datasource.content_url}, does not match the content URL of the uploaded datasource,{uploaded_datasource.content_url}.  This may cause errors in uploaded workbooks"
                    )

                os.remove(file_path)

                if datasource_name in list_to_add and (
                    datasource_options.get("include_published_datasources")
                    or not uploaded_datasource.has_extracts
                ):
                    logger.info(
                        f"Processing Datasource Connections for: {datasource_name} | {uploaded_datasource.has_extracts}"
                    )
                    self.update_connection_credentials(
                        uploaded_datasource, self.AssetType.Datasources
                    )
                    logger.info(
                        f"Processed Datasource Connections for: {datasource_name} | {uploaded_datasource.has_extracts}"
                    )
        logger.info(f"Added {len(to_add_and_update)} Datasources")

    def remove(self, remove_list: list):
        """
        Match the workbook from list to reference object
        Delete Workbook
        """
        if len(remove_list) > 0:
            for datasource_name in remove_list:
                datasource = self.get_item_from_list(
                    datasource_name, self.target_datasources
                )

                self.delete(self.plan.target, datasource)
                logger.info(f"Datasource {datasource.name} deleted")
        logger.info(f"Removed {len(remove_list)} Datasources")



"""Module for general Tableau Asset Manager"""
from TableauConMan.helpers import utils
from typing import Optional
import tableauserverclient as TSC


class AssetsManager:
    """

    :class: AssetsManager

    This class provides functionality for managing assets in a server.

    :class:`AssetType` defines different types of assets that can be managed using the `AssetsManager` class.
    """

    class AssetType:
        """

        :class: AssetType

        This class defines different types of assets.

        :class:`AssetType` has the following attributes:

        - `Workbooks`: Represents the asset type for workbooks.
        - `Datasources`: Represents the asset type for datasources.
        - `Groups`: Represents the asset type for groups.
        - `Projects`: Represents the asset type for projects.
        - `Users`: Represents the asset type for users.
        - `Workbook`: Represents the asset type for a single workbook.
        - `Project`: Represents the asset type for a single project.
        - `Datasource`: Represents the asset type for a single datasource.

        These attributes can be accessed using the `.` notation, like `AssetType.Workbooks`, `AssetType.Datasources`, etc.

        Example usage:

        .. code-block:: python

            # Accessing asset types
            workbook_type = AssetType.Workbooks
            datasource_type = AssetType.Datasources
            group_type = AssetType.Groups
            project_type = AssetType.Projects
            user_type = AssetType.Users
            single_workbook_type = AssetType.Workbook
            single_project_type = AssetType.Project
            single_datasource_type = AssetType.Datasource

            print(workbook_type)  # "workbooks"
            print(datasource_type)  # "datasources"
            print(group_type)  # "groups"
            print(project_type)  # "projects"
            print(user_type)  # "users"
            print(single_workbook_type)  # "workbook"
            print(single_project_type)  # "project"
            print(single_datasource_type)  # "datasource"

        """

        Workbooks = "workbooks"
        Datasources = "datasources"
        Groups = "groups"
        Projects = "projects"
        Users = "users"
        Workbook = "workbook"
        Project = "project"
        Datasource = "datasource"

    def __init__(self, plan) -> None:
        self.plan = plan

    @staticmethod
    def get_changes(list_a: list, list_b: list):
        """
        Compares the target and reference lists to get the differences
        """
        in_a_and_b, in_a_not_b, in_b_not_a = utils.compare_lists(list_a, list_b)

        return in_a_and_b, in_a_not_b, in_b_not_a

    @staticmethod
    def get_item_from_list(item_name, item_list):
        """
        Filters a list of items based on the name of the item
        """
        item = utils.get_item_from_list(item_name, item_list)

        return item

    def generate_server_list(
        self, source, asset_type, request_filter: Optional[TSC.RequestOptions] = None
    ):
        """

        :param source:
        :param asset_type:
        :param filter:
        :return:
        """
        asset = getattr(source.server, asset_type)

        with source.connect():
            asset_items = list(TSC.Pager(asset, request_filter))
            """
              Generalized: can be moved to parent class
              Find a way to included updated_at check.  Might be able to use dict to
              """

            asset_list = []
            for asset in asset_items:
                asset_list.append(asset.name)
            return asset_items, asset_list

    def update_connection_credentials(self, asset, asset_type):
        """
        Updates the connection credentials with a valid ses of credentials
        """
        # logger.info(f"Processing Datasource Connections for: {datasource_name} | {uploaded_datasource.has_extracts}")
        valid_connections = self.plan.connections
        target = self.plan.target
        server_asset = getattr(target.server, asset_type)
        with target.connect():
            server_asset.populate_connections(asset)

            for connection in asset.connections:
                for valid_connection in valid_connections:
                    if connection.connection_type == valid_connection.get(
                        "type"
                    ):  # In a workbook Published Datasource are of connection_type = sqlproxy
                        username, password = self.plan.get_connection_secretes(
                            valid_connection
                        )

                        connection.username = username
                        connection.password = password
                        connection.embed_password = True

                        server_asset.update_connection(asset, connection)



from enum import Enum

INCLUDE_ASSET_PERMISSIONS = False
DELETE_ON_TARGET = False


class ConfigureAssets(Enum):
    # Sites = "sites"
    # Users = "users"
    # Groups = "groups"
    Projects = "projects"



from loguru import logger
from jinja2 import Environment, BaseLoader, FileSystemLoader
import os


def envvar(variable_name):
    return os.environ.get(variable_name)


def get_jinja_environment(template_path=None):
    if template_path:
        templateLoader = FileSystemLoader(searchpath=template_path)
    else:
        templateLoader = BaseLoader()

    jinja_environment = Environment(loader=templateLoader)
    jinja_environment.globals.update(envvar=envvar)

    return jinja_environment


def compare_lists(list_a: list, list_b: list):
    """
    This function will take two lists and compare them retuning
    three lists for the list overlaps and the unique values in each list
    """

    in_a_and_b = sorted(list(set(list_a) & set(list_b)))
    in_a_not_b = sorted(list(set(list_a) ^ set(in_a_and_b)))
    in_b_not_a = sorted(list(set(list_b) ^ set(in_a_and_b)))

    return in_a_and_b, in_a_not_b, in_b_not_a


def get_item_from_list(item_name: str, item_list: list):
    """
    Filters a list of items based on the name of the item
    """
    try:
        item = list(filter(lambda x: item_name == x.name, item_list))[0]

    except Exception as exc:
        logger.info(f"Could not find group: {item_name} in the item list")
        logger.error(exc)

    return item


def get_filtered_item_list(
    item_list: list, attribute_list: list, filter_attribute: str
):
    """
    Filters a list of Items based on a specific list of values
    """

    filtered_list = list(
        filter(lambda x: getattr(x, filter_attribute) in attribute_list, item_list)
    )

    return filtered_list


def get_filtered_dict_list(dict_list: list, key_list: list, filter_key: str):
    """
    Filters a list of dictionaries based on a specific list of keys
    """
    filter_to = list()
    if not isinstance(key_list, list):
        filter_to.append(key_list)
    else:
        filter_to = key_list

    filtered_list = list(filter(lambda x: x.get(filter_key) in filter_to, dict_list))

    return filtered_list


def get_list_of_attr(source_list: list, target_attr: str):
    result_list = []
    for element in source_list:
        result_list.append(element.get(target_attr))
    return result_list



from dataclasses import dataclass, field
from typing import List

from TableauConMan.assets.asset import Asset
from TableauConMan.assets.permission_rule_asset import PermissionRule
import hashlib
import json


@dataclass
class Project(Asset):
    project_id: str
    project_path: str
    project_parent_id: str
    project_name: str
    description: str
    content_permissions: str
    parent_content_permissions: str
    permission_set: List[PermissionRule] = field(default_factory=list)

    @property
    def manage_permissions(self) -> bool:
        return (
            self.parent_content_permissions != "LockedToProject"
            and self.content_permissions != "ManagedByOwner"
        )

    def update_asset(self, reference_asset: Asset) -> None:
        self.description = reference_asset.description
        self.content_permissions = reference_asset.content_permissions

    @property
    def specification_id(self):
        return hashlib.md5(self.project_path.encode("utf-8")).hexdigest()

    @property
    def update_checksum(self):
        return hashlib.md5(
            (
                self.specification_id
                + self.description
                + self.content_permissions
                + json.dumps(
                    self.permission_set, default=lambda o: o.__dict__, sort_keys=True
                )
            ).encode("utf-8")
        ).hexdigest()

    @property
    def update_value(self):
        return 1



from typing import Optional, Type, List
from dataclasses import dataclass, field
from loguru import logger
from TableauConMan.sources.sources import Source
from TableauConMan.factories.source_factory import SourceFactory


@dataclass
class Plan:
    target: Optional[Source] = None
    reference: Source = None
    assets: list = field(default_factory=list)
    _options: list = None

    @property
    def project_options(self) -> List[dict]:
        """_summary_

        :return: _description_
        """
        return self._options.get("projects", {}).get("options")

    @classmethod
    def load(cls: Type["Plan"], plan_file: str):
        """_summary_

        :param cls: _description_
        :param plan_file: _description_
        :return: _description_
        """
        valid_plan = cls.validate_plan(plan_file)

        source_factory = SourceFactory()

        target_source = source_factory.get_source(
            valid_plan.get("target", {}).get("type")
        )

        reference_source = source_factory.get_source(
            valid_plan.get("reference", {}).get("type")
        )

        return cls(
            target=target_source.load(valid_plan.get("target")),
            reference=reference_source.load(valid_plan.get("reference")),
            assets=list(valid_plan.get("assets").keys()),
            _options=valid_plan.get("assets"),
        )

    @staticmethod
    def validate_plan(plan_file):
        """_summary_

        :param plan_file: _description_
        :return: _description_
        """
        return plan_file



"""Module management of provided plans"""
import os

import tableauserverclient as TSC
from TableauConMan.server_connector import ServerConnector
from TableauConMan.yaml_connector import YamlConnector
from TableauConMan.specification import Specification
from loguru import logger


class Plan:
    """
    Plan class represents a plan for data manipulation in a system.

    Attributes:
        target (str): The target connection for the plan.
        reference (str): The reference connection for the plan.
        target_selection_rules (TSC.RequestOptions): The selection rules for the target connection.
        reference_selection_rules (TSC.RequestOptions): The selection rules for the reference connection.
        assets (list): The assets to be manipulated by the plan.
        asset_options (dict): The options for the assets.
        operation (str): The operation to be performed by the plan.
        connections (dict): The connections used by the plan.
        raw_plan (dict): The raw plan data.

    Note:
        - The `target` and `reference` attributes represent connection objects.
        - The selection rules are instances of the `TSC.RequestOptions` class.
        - The assets are represented by a list.
        - The asset options are represented by a dictionary.
        - The operation is represented by a string.
        - The connections are represented by a dictionary.
        - The raw plan is represented by a dictionary.
    """

    def __init__(self):
        self.target: str = None
        self.reference: str = None
        self.target_selection_rules: TSC.RequestOptions() = None
        self.reference_selection_rules: TSC.RequestOptions() = None
        self.assets: list = None
        self.asset_options: dict = None
        self.operation: str = None
        self.connections: dict = None
        self.raw_plan: dict = None

    def load_plan(self, raw_plan: dict) -> None:
        """

        :param raw_plan:
        """
        if not isinstance(raw_plan, dict):
            raise TypeError("The raw plan must be a dictionary")

        self.target = self.format_connection(raw_plan.get("target"))
        self.reference = self.format_connection(raw_plan.get("reference"))
        self.target_selection_rules = self.format_rule(
            raw_plan.get("target_selection_rules")
        )
        self.reference_selection_rules = self.format_rule(
            raw_plan.get("reference_selection_rules")
        )
        self.assets = list(raw_plan.get("assets"))
        self.asset_options = raw_plan.get("assets")
        self.operation = raw_plan.get("operation")
        self.connections = raw_plan.get("connections")
        self.raw_plan = raw_plan

    @staticmethod
    def format_rule(rules):
        """

        :param rules:
        :return:
        """
        formatted_rule = TSC.RequestOptions()

        if rules:
            for rule in rules:
                field = getattr(formatted_rule.Field, rule.get("field"))
                operator = getattr(formatted_rule.Operator, rule.get("operator"))
                value = rule.get("value")
                formatted_rule.filter.add(TSC.Filter(field, operator, value))

        return formatted_rule

    def format_connection(self, connection):
        """

        :param connection:
        :return:
        """
        connection_type = connection.get("type")

        if connection_type == "server":
            connection = self.format_server(connection)
        elif connection_type == "spec":
            connection = self.format_spec(connection)

        return connection

    def format_server(self, connection):
        """

        :param connection:
        :return:
        """
        (
            token_name,
            token_secret,
            server_url,
            site_name,
        ) = self.get_server_secrets_from_env(connection)

        server = ServerConnector(site_name, server_url, token_name, token_secret)

        return server

    def format_spec(self, connection):
        """

        :param connection:
        :return:
        """
        spec_file_path = connection.get("file_path")

        raw_spec = YamlConnector(spec_file_path)

        spec = Specification()
        spec.load_spec(raw_spec.get_yaml())

        return spec

    @staticmethod
    def get_server_secrets(server):
        """

        :param server:
        :return:
        """
        token_name = server.get("token_name")
        token_secret = server.get("token_secret")
        server_url = server.get("server_url")
        site_name = server.get("site_name")

        return token_name, token_secret, server_url, site_name

    def get_server_secrets_from_env(self, server):
        """

        :param server:
        :return:
        """
        token_name = os.environ.get(server.get("token_name")).strip()
        token_secret = os.environ.get(server.get("token_secret")).strip()
        server_url = os.environ.get(server.get("server_url")).strip()
        site_name = os.environ.get(server.get("site_name")).strip()

        return token_name, token_secret, server_url, site_name

    @staticmethod
    def get_connection_secretes(connection):
        """

        :param connection:
        :return:
        """

        username = os.environ.get(connection.get("username")).strip()
        password = os.environ.get(connection.get("password")).strip()

        return username, password



import os
from typing import Any, List, Optional, Type
from dataclasses import dataclass

import tableauserverclient as TSC
from loguru import logger


@dataclass
class ServerSource:
    """Class representing the connection details and methods for connecting to a Tableau server"""

    def __init__(
        self,
        site_name: Optional[str] = None,
        server_url: Optional[str] = None,
        token_name: Optional[str] = None,
        token_secret: Optional[str] = None,
    ):
        self.site_name: Optional[str] = site_name
        self.server_url: Optional[str] = server_url
        self.token_name: Optional[str] = token_name
        self.token_secret: Optional[str] = token_secret
        self.authentication: Optional[TSC.PersonalAccessTokenAuth] = None
        self.server: Optional[TSC.Server] = None
        self.source_type: str = "server"

    """def __repr__(self):

        return "<Server {site_name} server_object={_server} auth_object={_authentication}>".format(
            **self.__dict__
        )"""

    @property
    def name(self) -> str:
        return self.site_name

    @property
    def server(self) -> Optional[TSC.Server]:
        """

        :return:
        """
        return self._server

    @server.setter
    def server(self, value):
        if value is None:
            self._server = TSC.Server(self.server_url, use_server_version=True)
        elif not isinstance(value, TSC.Server):
            raise TypeError("server must be of TSC.Server class")
        else:
            self._server = value

    @property
    def authentication(self) -> TSC.PersonalAccessTokenAuth:
        """

        :return:
        """
        return self._authentication

    @authentication.setter
    def authentication(self, value):
        if value is None:
            self._authentication = TSC.PersonalAccessTokenAuth(
                self.token_name,
                self.token_secret,
                site_id=self.site_name,
            )
        elif not isinstance(value, TSC.PersonalAccessTokenAuth):
            raise TypeError(
                "authentication must be of  TSC.PersonalAccessTokenAuth class"
            )
        else:
            self._authentication = value

    def connect(self):
        """

        :return:
        """
        connection = self.server.auth.sign_in(self.authentication)

        return connection

    def get_assets(
        self,
        asset_type: str,
        request_filter: Optional[TSC.RequestOptions] = None,
    ) -> List[Any]:
        """

        :param source:
        :param asset_type:
        :param filter:
        :return:
        """
        asset = getattr(self.server, asset_type)

        with self.connect():
            asset_items = list(TSC.Pager(asset, request_filter))
            return asset_items

    def delete_item(self, asset_type: str, item_to_delete):
        server_item = getattr(self.server, asset_type)

        with self.connect():
            server_item.delete(item_to_delete.id)

    def add_item(self, asset_type: str, item_to_delete):
        server_item = getattr(self.server, asset_type)

        with self.connect():
            new_item = server_item.create(item_to_delete)

        return new_item

    def update_item(self, asset_type: str, item_to_delete):
        server_item = getattr(self.server, asset_type)

        with self.connect():
            updated_item = server_item.update(item_to_delete)

        return updated_item

    def add_sub_item(self, asset_type: str, target_item, sub_item) -> None:
        target_item = self.populate_default_permissions(target_item, asset_type)
        with self.connect():
            server_add_function = self._get_project_permission_method(
                asset_type, "update"
            )
            # TSC is expecting a list of rules.
            server_add_function(target_item, [sub_item])

    def remove_sub_item(self, asset_type: str, target_item, sub_item) -> None:
        target_item = self.populate_default_permissions(target_item, asset_type)
        with self.connect():
            server_add_function = self._get_project_permission_method(
                asset_type, "delete"
            )
            server_add_function(target_item, sub_item)

    def update_sub_item(self):
        ...

    def populate_default_permissions(
        self, project_item: TSC.ProjectItem, asset_type: str
    ) -> TSC.ProjectItem:
        populate_method = self._get_project_permission_method(asset_type, "populate")

        with self.connect():
            populate_method(project_item)
            return project_item

    def _get_project_permission_method(self, asset_type, action):
        if asset_type == "project":
            if action == "update":
                permission_method_name = "update_permission"
            elif action == "delete":
                permission_method_name = "delete_permission"
            elif action == "populate":
                permission_method_name = "populate_permissions"
        else:
            if action == "update":
                permission_method_name = f"update_{asset_type}_default_permissions"
            elif action == "delete":
                permission_method_name = f"delete_{asset_type}_default_permissions"
            elif action == "populate":
                permission_method_name = f"populate_{asset_type}_default_permissions"

        permission_method = getattr(self.server.projects, permission_method_name)

        return permission_method

    def _get_project_permission_object(self, asset_type, project_item):
        if asset_type == "project":
            permission_object_name = "permissions"
        else:
            permission_object_name = f"default_{asset_type}_permissions"

        permission_object = getattr(project_item, permission_object_name)

        return permission_object

    @classmethod
    def load(cls: Type["ServerSource"], source_config: dict):
        return cls(
            token_name=os.environ.get(source_config.get("token_name")).strip(),
            token_secret=os.environ.get(source_config.get("token_secret")).strip(),
            server_url=os.environ.get(source_config.get("server_url")).strip(),
            site_name=os.environ.get(source_config.get("site_name")).strip(),
        )



"""
This file describes the expected schema for a spec file.
These schemas are used to both parse and validate spec files.
"""


SPEC_SCHEMA = """
    version:
        type: string
        required: False

    site:
        type: dict
        required: False
        keyschema:
            type: string


    projects:
        type: list
        schema:
            type: dict
            keyschema:
                type: string
            valuesrules:
                anyof:
                  - type: string
                  - type: list

    permission_templates:
        type: list
        schema:
            type: dict
            keyschema:
                type: string
            valuesrules:
                anyof:
                  - type: string
                  - type: list
                  - type: dict
    groups:
        type: list
        schema:
            type: dict
            keyschema:
                type: string
            valuesrules:
                type: string
    users:
        type: list
        schema:
            type: dict
            keyschema:
                type: string
            valuesrules:
                anyof:
                  - type: string
                  - type: list

"""

SPEC_PROJECTS_SCHEMA = """
    project_name:
        type: string

    project_id:
        type: string

    description:
        type: string

    content_permissions:
        type: string

    project_path:
        type: string

    permission_set:
        type: list
        schema:
            type: dict
            allowed:
                - group_name
                - user_name
                - permission_template_id
                - permission_rule
            keyschema:
                type: string
            valuesrules:
                type: string
            
"""

SPEC_PERMISSION_TEMPLATES_SCHEMA = """
    name:
      type: string

    permission_template_id:
      type: string

    workbook:
      type: dict
      allowed:
          - Read
          - Filter
          - ViewComments
          - AddComment
          - ExportImage
          - ExportData
          - ShareView
          - ViewUnderlyingData
          - WebAuthoring
          - RunExplainData
          - ExportXml
          - Write
          - CreateRefreshMetrics
          - ChangeHierarchy
          - ChangePermissions
          - Delete
      keyschema:
          type: string
      valuesrules:
          type: string
          allowed:
              - Allow
              - Deny

    project:
      type: dict
      allowed:
          - Read
          - Write
          - ProjectLeader
          - InheritedProjectLeader
      keyschema:
          type: string
      valuesrules:
          type: string
          allowed:
              - Allow
              - Deny

    datasource:
        type: dict
        allowed:
            - Read
            - Connect
            - ExportXml
            - Write
            - SaveAs
            - ChangeHierarchy
            - ChangePermissions
            - Delete
        keyschema:
          type: string
        valuesrules:
            type: string
            allowed:
                - Allow
                - Deny

    flow:
        type: dict
        allowed:
            - Read
            - ExportXml
            - Execute
            - Write
            - WebAuthoringForFlows
            - ChangeHierarchy
            - ChangePermissions
            - Delete
        keyschema:
          type: string
        valuesrules:
            type: string
            allowed:
                - Allow
                - Deny

    datarole:
        type: dict
        allowed:
            - Read
            - Write
            - ChangeHierarchy
            - ChangePermissions
            - Delete
        keyschema:
          type: string
        valuesrules:
            type: string
            allowed:
                - Allow
                - Deny

    lens:
        type: dict
        allowed:
            - Read
            - Write
            - ChangeHierarchy
            - ChangePermissions
            - Delete
        keyschema:
          type: string
        valuesrules:
            type: string
            allowed:
                - Allow
                - Deny

    metric:
        type: dict
        allowed:
            - Read
            - Write
            - ChangeHierarchy
            - ChangePermissions
            - Delete
        keyschema:
          type: string
        valuesrules:
            type: string
            allowed:
                - Allow
                - Deny

    virtual_connection:
        type: dict
        allowed:
            - Read
            - Connect
            - Write
            - ChangeHierarchy
            - ChangePermissions
            - Delete
        keyschema:
          type: string
        valuesrules:
            type: string
            allowed:
                - Allow
                - Deny

    database:
        type: dict
        allowed:
            - Read
            - Write
            - ChangeHierarchy
            - ChangePermissions
        keyschema:
          type: string
        valuesrules:
            type: string
            allowed:
                - Allow
                - Deny

    table:
        type: dict
        allowed:
            - Read
            - Write
            - ChangeHierarchy
            - ChangePermissions
        keyschema:
          type: string
        valuesrules:
            type: string
            allowed:
                - Allow
                - Deny



"""

SPEC_GROUPS_SCHEMA = """

    group_name:
        type: string
"""

SPEC_USERS_SCHEMA = """

    user_name: 
        type: string

    user_name_domain:
        type: string

    site_role:
        type: string
        allowed:
            - SiteAdministratorCreator
            - Creator
            - SiteAdministratorExplorer
            - ExplorerCanPublish
            - Explorer
            - Viewer
            - Unlicensed
    
    auth_setting:
        type: string
        allowed:
            - SAML
            - TableauIDWithMFA

    groups:
        type: list
"""



from typing import Any
from TableauConMan.managers.project_manager import ProjectManager


class AssetManagerFactory:
    @staticmethod
    def get_asset_manager(asset_type: str) -> Any:
        if asset_type == "projects":
            return ProjectManager()
        raise TypeError(f"No manager for the {asset_type} asset_type")



from typing import Optional
from TableauConMan.assets_manager import AssetsManager
from TableauConMan.helpers import utils
import tableauserverclient as TSC
from loguru import logger


class GroupsManager(AssetsManager):
    """

    :class: `GroupsManager` is a subclass of `AssetsManager` that provides functionality for managing groups in a Tableau Server. It includes methods for populating groups, generating a
    * server list, getting group changes, deleting a group, creating a group, adding and removing users from a group, and updating group membership.

    :class: `GroupsManager` inherits from `AssetsManager`. The `AssetsManager` class provides basic functionality for managing assets in a Tableau Server, such as workbooks, data sources
    *, and projects.

    :param plan: A Plan object that contains information about the action to be performed.

    :ivar target_groups: A list of `TSC.GroupItem` objects representing the groups in the target server.
    :ivar target_groups_list: A list of group names in the target server.
    :ivar target_group_memberships_list: A list of dictionaries representing group memberships in the target server.
    :ivar reference_groups: A list of `TSC.GroupItem` objects representing the groups in the reference server.
    :ivar reference_groups_list: A list of group names in the reference server.
    :ivar reference_group_memberships_list: A list of dictionaries representing group memberships in the reference server.
    :ivar target_users: A list of user names in the target server.
    :ivar reference_users: A list of user names in the reference server.

    Methods:

    .. method:: populate_groups()

       Use object properties to get a comparable list of groups from the source and reference servers.

    .. method:: populate_users()

       Use object properties to get a list of users from the target server.

    .. method:: _generate_server_list(source, asset_type, filter=None)

       Generate a list of assets from a server.

       :param source: The server to generate the list from.
       :param asset_type: The type of asset to generate the list for.
       :param filter: Optional. A filter to apply to the asset list.

       :return: A tuple containing the list of assets and the list of asset names.

    .. method:: get_group_changes()

       Get the changes between the target and reference groups.

       :return: A tuple containing the groups to be added, the groups to be removed, and the groups to be updated.

    .. method:: get_group_memberships(group_name)

       Get the memberships of a specific group.

       :param group_name: The name of the group.

       :return: A tuple containing the reference memberships and the target memberships.

    .. method:: delete_group(group)

       Delete a group from the target server.

       :param group: The group to delete.

    .. method:: create(group_name)

       Create a new group in the target server.

       :param group_name: The name of the group to create.

       :return: The created group.

    .. method:: add_user(group_item, user_item)

       Add a user to a group.

       :param group_item: The group to add the user to.
       :param user_item: The user to add to the group.

    .. method:: remove_user(group_item, user_item)

       Remove a user from a group.

       :param group_item: The group to remove the user from.
       :param user_item: The user to remove from the group.

    .. method:: update_group_membership(group_item)

       Update the membership of a group.

       :param group_item: The group to update the membership of.

    .. method:: add(list_to_add)

       Add groups to the target server.

       :param list_to_add: A list of group names to add.

    .. method:: remove(remove_list)

       Remove groups from the target server.

       :param remove_list: A list of group names to remove.

    .. method:: update(update_list)

       Update groups in the target server.

       :param update_list: A list of group names to update.

    """

    def __init__(self, plan) -> None:
        """

        :param plan:
        """
        AssetsManager.__init__(self, plan)
        self.target_groups: Optional[list[TSC.GroupItem]] = None
        self.target_groups_list: Optional[list] = None
        self.target_group_memberships_list: Optional[list] = None
        self.reference_groups: Optional[list[TSC.GroupItem]] = None
        self.reference_groups_list: Optional[list] = None
        self.reference_group_memberships_list: Optional[list] = None
        self.target_users: Optional[list] = None
        self.reference_users: Optional[list] = None

    def populate_groups(self):
        """
        Use object properties to get a comparable list of workbooks from the source and reference
        """

        full_target_groups, full_target_groups_list = self.generate_server_list(
            self.plan.target, self.AssetType.Groups
        )

        self.target_groups_list = list(
            filter(lambda x: x != "All Users", full_target_groups_list)
        )

        self.target_groups = list(
            filter(lambda x: x.name != "All Users", full_target_groups)
        )

        group_memberships = list()

        for group in self.target_groups:
            with self.plan.target.connect():
                self.plan.target.server.groups.populate_users(group)

                group_membership_dict = {"group_name": group.name, "users": []}

                for user in group.users:
                    group_membership_dict.get("users").append(user.name)

                group_memberships.append(group_membership_dict)

        self.target_group_memberships_list = group_memberships

        # Load the spec as part of the Plan?
        """plan_reference = self.plan.raw_plan.get("reference")
        spec_file_path = plan_reference.get("file_path")

        raw_spec = YamlConnector(spec_file_path)

        spec = Specification()
        spec.load_spec(raw_spec.get_yaml())"""

        spec = self.plan.reference

        filtered_groups = list(
            filter(lambda x: x.get("group_name") != "All Users", spec.groups)
        )

        self.reference_groups_list = list(x.get("group_name") for x in filtered_groups)

        self.reference_group_memberships_list = filtered_groups.copy()

        for group in self.reference_group_memberships_list:
            group.update({"users": []})
            for user in spec.users:
                if user.get("user_name_domain"):
                    full_user_name = (
                        user.get("user_name") + "@" + user.get("user_name_domain")
                    )
                else:
                    full_user_name = user.get("user_name")
                if "groups" in user:
                    for user_group in user.get("groups"):
                        if user_group == group.get("group_name"):
                            group.get("users").append(full_user_name)

                        if user_group not in (
                            x.get("group_name")
                            for x in self.reference_group_memberships_list
                        ):
                            logger.warning(
                                f"An exception occurred: Group {user_group} on user {full_user_name} is not valid"
                            )

    def populate_users(self):
        full_target_users, full_target_users_list = self.generate_server_list(
            self.plan.target, self.AssetType.Users
        )

        self.target_users = full_target_users

    def _generate_server_list(
        self, source, asset_type, request_filter: Optional[TSC.RequestOptions] = None
    ):
        """

        :param source:
        :param asset_type:
        :param filter:
        :return:
        """
        asset = getattr(source.server, asset_type)

        with source.connect():
            asset_items = list(TSC.Pager(asset, request_filter))
            """
              Generalized: can be moved to parent class
              Find a way to included updated_at check.  Might be able to use dict to
              """

            asset_list = []
            for asset in asset_items:
                asset_list.append(asset.name)
            return asset_items, asset_list

    def get_group_changes(self):
        """

        :return:
        """
        common, to_remove, to_add = self.get_changes(
            self.target_groups_list, self.reference_groups_list
        )

        update_list = list()

        for group_name in common:
            reference_membership, target_membership = self.get_group_memberships(
                group_name
            )

            logger.debug(
                f"Group: {group_name} | Target: {set(target_membership)} | Reference: {set(reference_membership)} | Check: {set(target_membership) != set(reference_membership)}"
            )

            if set(target_membership) != set(reference_membership):
                update_list.append(group_name)

        return to_add, to_remove, update_list

    def get_group_memberships(self, group_name):
        reference_membership_group = utils.get_filtered_dict_list(
            self.reference_group_memberships_list, group_name, "group_name"
        )[0]
        reference_membership = reference_membership_group.get("users")

        try:
            target_membership_group = utils.get_filtered_dict_list(
                self.target_group_memberships_list, group_name, "group_name"
            )[0]

            target_membership = target_membership_group.get("users")
        except:
            target_membership = list()

        logger.debug(
            f"Group: {group_name} | Reference Memberships: {self.reference_group_memberships_list} | Target Memberships: {self.target_group_memberships_list} | Reference Users: {reference_membership} | Target Users: {target_membership} "
        )

        return reference_membership, target_membership

    def delete_group(self, group):
        with self.plan.target.connect():
            self.plan.target.server.groups.delete(group.id)
            logger.info(f"Group {group.name} was removed from the server")

    def create(self, group_name):
        # create a new instance with the group name
        new_group = TSC.GroupItem(group_name)

        with self.plan.target.connect():
            created_group = self.plan.target.server.groups.create(new_group)
            logger.info(f"Group {created_group} was added to the server")

        return created_group

    def add_user(self, group_item, user_item):
        with self.plan.target.connect():
            self.plan.target.server.groups.add_user(group_item, user_item.id)
        logger.info(f"User {user_item.name} was added to group {group_item.name}")

    def remove_user(self, group_item, user_item):
        with self.plan.target.connect():
            self.plan.target.server.groups.remove_user(group_item, user_item.id)
        logger.info(f"User {user_item.name} was removed from group {group_item.name}")

    def update_group_membership(self, group_item):
        group_name = group_item.name
        reference_membership, target_membership = self.get_group_memberships(group_name)

        common, to_remove, to_add = self.get_changes(
            target_membership, reference_membership
        )

        logger.debug(f"Group: {group_name}, Add: {to_add}, Remove: {to_remove}")

        for user in to_add:
            user_item = utils.get_item_from_list(user, self.target_users)
            self.add_user(group_item, user_item)

        for user in to_remove:
            user_item = utils.get_item_from_list(user, self.target_users)
            self.remove_user(group_item, user_item)

        logger.info(f"Added {len(to_add)} users and removed {len(to_remove)} users")

    def add(self, list_to_add: list):
        """ """

        if len(list_to_add) > 0:
            for group_name in list_to_add:
                logger.info(f"Processing the addition of Group: {group_name}")

                created_group = self.create(group_name)

                self.update_group_membership(created_group)

        logger.info(f"Added {len(list_to_add)} Groups")

    def remove(self, remove_list: list):
        """
        Match the workbook from list to reference object
        Delete Workbook
        """
        if len(remove_list) > 0:
            for group_name in remove_list:
                group = self.get_item_from_list(group_name, self.target_groups)
                logger.debug(f"Group to Remove: {group}")

                self.delete_group(group)
        logger.info(f"Removed {len(remove_list)} Groups")

    def update(self, update_list: list):
        """ """
        if len(update_list) > 0:
            for group_name in update_list:
                group_item = self.get_item_from_list(group_name, self.target_groups)

                self.update_group_membership(group_item)
        logger.info(f"Updated {len(update_list)} Groups")



from typing import Any, List, Type
import os

from dataclasses import dataclass
import yaml
import hashlib
import json
from TableauConMan.helpers import utils

from loguru import logger


class IndentDumper(yaml.Dumper):
    def increase_indent(self, flow=False, indentless=False):
        return super(IndentDumper, self).increase_indent(flow, False)


@dataclass
class SpecificationSource:
    """Class representing a connection to a yaml file or yaml string"""

    file_path: str = None
    source_type: str = "specification"

    @property
    def name(self) -> str:
        return self.file_path

    @property
    def file_text_list(self):
        return self.read_yaml()

    def read_yaml(self):
        """

        :param object:
        :return:
        """

        file_path_parts = os.path.split(self.file_path)
        search_path = file_path_parts[0]
        file_name = file_path_parts[1]
        # logger.debug(f"{os.getcwd()} | {search_path}: {file_name}")
        jinja_environment = utils.get_jinja_environment(search_path)

        try:
            file_text_list = yaml.safe_load(
                jinja_environment.get_template(file_name).render()
            )
            return file_text_list
        except yaml.YAMLError as exc:
            print(exc)
            return None

    def write_yaml(self, dict_file, filename):
        with open(f"{filename}", "w") as file:
            documents = yaml.dump(
                dict_file,
                file,
                default_flow_style=False,
                sort_keys=False,
                Dumper=IndentDumper,
            )

    def get_assets(self, asset_type: str) -> List[Any]:
        asset_list = self.file_text_list.get(asset_type)

        if asset_type == "projects":
            asset_list = self.read_permission_rules(asset_list)

        return asset_list

    def get_permission_template(self, project_permission: dict) -> dict:
        permission_templates = self.get_assets("permission_templates")
        # logger.debug(f"{permission_templates} | {project_permission}")
        template_name = project_permission.get("permission_rule")
        project_permission_rule = next(
            item for item in permission_templates if item.get("name") == template_name
        ).copy()

        project_permission_rule.pop("name")

        return project_permission_rule

    def read_permission_rules(self, project_specification_list: dict) -> List[dict]:
        for project_specification in project_specification_list:
            if project_specification.get("permission_set"):
                for permission in project_specification.get("permission_set"):
                    rule = self.get_permission_template(permission)

                    permission.update({"permission_rule": rule})

        return project_specification_list

    def write_permission_templates(
        self, file_text: dict, project_list: List[dict]
    ) -> dict:
        permission_templates_list = list()

        for project in project_list:
            for permission_rule in project.get("permission_set"):
                permission_template = permission_rule.copy()
                permission_template_id = hashlib.md5(
                    json.dumps(permission_template, sort_keys=True).encode("utf-8")
                ).hexdigest()

                permission_rule.update({"permission_rule": permission_template_id})

                permission_template.update({"name": permission_template_id})

                for asset, capabilities in permission_template.get(
                    "permission_rule"
                ).items():
                    permission_template.update({asset: capabilities})

                permission_template.pop("permission_rule")

                if permission_template.get("group_name"):
                    permission_template.pop("group_name")

                if permission_template.get("user_name"):
                    permission_template.pop("user_name")

                permission_templates_list.append(permission_template)

        unique_permission_templates = list(
            {v.get("name"): v for v in permission_templates_list}.values()
        )

        file_text.update({"permission_templates": unique_permission_templates})

        return file_text, project_list

    def set_asset(self, asset_type: str, asset_list: List[dict]) -> None:
        current_file = self.file_text_list

        if asset_type == "projects":
            current_file, asset_list = self.write_permission_templates(
                current_file, asset_list
            )

        current_file.update({asset_type: asset_list})

        self.write_yaml(current_file, self.file_path)

    @classmethod
    def load(cls: Type["SpecificationSource"], source_config: dict):
        return cls(file_path=source_config.get("file_path"))



import click
import TableauConMan
from loguru import logger


@click.group(invoke_without_command=True, no_args_is_help=True)
@click.version_option(version=TableauConMan.__version__, prog_name="TableauConMan")
@click.pass_context
def cli(ctx):
    ctx.ensure_object(dict)



""" Abstract class defining the structure of a converter
"""
from abc import ABC, abstractmethod
from TableauConMan.assets.asset import Asset


class Converter(ABC):
    """Methods for converting between TSC Items and ConMan Assets"""

    @abstractmethod
    def convert_to_asset(self, source_item) -> Asset:
        """factory method for converting from a source"""

    @abstractmethod
    def convert_to_source(self, asset):
        """factory method for converting to a source"""



from TableauConMan.sources.server_source import ServerSource
from TableauConMan.sources.specification_source import SpecificationSource


class SourceFactory:
    def get_source(self, source_type_config: str):
        if source_type_config == "server":
            return ServerSource
        elif source_type_config == "specification":
            return SpecificationSource
        raise TypeError(f"No source for the {source_type_config} source type")



from abc import ABC, abstractmethod
from TableauConMan.sources.sources import Source
from TableauConMan.assets.asset import Asset
from TableauConMan.factories.converter_factory import ConverterFactory
import TableauConMan.helpers.utils as utils


from typing import List
from loguru import logger


class AssetManager(ABC):
    def get_asset_list(self, asset_type: str, asset_source: Source) -> List[Asset]:
        asset_list = list()

        item_list = asset_source.get_assets(asset_type)

        converter_generator = ConverterFactory()

        asset_converter = converter_generator.get_asset_converter(
            asset_type, asset_source
        )

        for item in item_list:
            asset_spec = asset_converter.convert_to_asset(item)
            asset_list.append(asset_spec)

        return asset_list

    def get_source_list(
        self, asset_type: str, asset_list: List[Asset], asset_source: Source
    ) -> List[Asset]:
        source_item_list = list()

        converter_generator = ConverterFactory()

        asset_converter = converter_generator.get_asset_converter(
            asset_type, asset_source
        )

        for asset in asset_list:
            source_item = asset_converter.convert_to_source(asset)
            source_item_list.append(source_item)

        return source_item_list

    def asset_specification_ids(self, asset_list: List[Asset]) -> List[str]:
        return [item.specification_id for item in asset_list]

    def group_asset_list(
        self, target_asset_list: List[Asset], reference_asset_list: List[Asset]
    ) -> (List[Asset], List[Asset], List[Asset]):
        reference_assets_ids = self.asset_specification_ids(reference_asset_list)
        target_assets_ids = self.asset_specification_ids(target_asset_list)

        in_both, in_target_not_reference, in_reference_not_target = utils.compare_lists(
            target_assets_ids, reference_assets_ids
        )

        asset_to_add = [
            item
            for item in reference_asset_list
            if item.specification_id in in_reference_not_target
        ]

        assets_to_remove = [
            item
            for item in target_asset_list
            if item.specification_id in in_target_not_reference
        ]

        in_both = [
            item for item in reference_asset_list if item.specification_id in in_both
        ]

        assets_to_update = self.check_for_update(
            in_both, target_asset_list, reference_asset_list
        )

        return asset_to_add, assets_to_remove, assets_to_update

    def check_for_update(
        self,
        asset_list: List[Asset],
        target_asset_list: List[Asset],
        reference_asset_list: List[Asset],
    ) -> List[Asset]:
        assets_to_update = list()

        for asset in asset_list:
            # logger.debug(f"Asset: {asset} | Asset List: {target_asset_list}")
            target_asset = next(
                item
                for item in target_asset_list
                if item.specification_id == asset.specification_id
            )

            reference_asset = next(
                item
                for item in reference_asset_list
                if item.specification_id == asset.specification_id
            )
            """logger.debug(
                f"Target asset: {target_asset} Reference asset: {reference_asset}"
            )"""
            if (
                target_asset.update_checksum != reference_asset.update_checksum
                or target_asset.update_value < reference_asset.update_value
            ):
                assets_to_update.append(tuple((target_asset, reference_asset)))

                # logger.debug(f"Added item for update: {assets_to_update}")

        return assets_to_update

    @abstractmethod
    def add(asset_type: str, asset_list: List[Asset], target_source: Source) -> None:
        pass

    @abstractmethod
    def remove(asset_type: str, asset_list: List[Asset], target_source: Source) -> None:
        pass

    @abstractmethod
    def update(asset_type: str, asset_list: List[Asset], target_source: Source) -> None:
        pass



from typing import Optional, Dict, List
from TableauConMan.assets_manager import AssetsManager
from TableauConMan.helpers import utils
import tableauserverclient as TSC
from loguru import logger


class UserManager(AssetsManager):
    def __init__(self, plan) -> None:
        """

        :param plan:
        """
        AssetsManager.__init__(self, plan)
        self.target_groups: Optional[list[TSC.GroupItem]] = None
        self.target_groups_list: Optional[list] = None
        self.target_group_memberships_list: Optional[list] = None
        self.reference_groups: Optional[list[TSC.GroupItem]] = None
        self.reference_groups_list: Optional[list] = None
        self.reference_group_memberships_list: Optional[list] = None
        self.target_users: Optional[list] = None
        self.reference_users: Optional[list] = None

    def populate_server_users(self) -> Dict:
        """

        :return:
        """
        logger.info("Retrieving server users")
        with self.plan.target.connect():
            clean_dict = {
                user.email: user.site_role
                for user in TSC.Pager(self.plan.target.server.users)
            }
        logger.success("Server users retrieved")
        return clean_dict

    def create_user(self, user_name: str, role: str):
        """

        :param user_name:
        :param role:
        :return:
        """
        logger.info(f"Creating user: {user_name}, role: {role}")
        new_user = TSC.UserItem(user_name, role, auth_setting="SAML")

        with self.plan.target.connect():
            new_u = self.plan.target.server.users.add(new_user)
            logger.success("User added successfully")
            return new_u.id

    def delete_user(self, user_name: str) -> bool:
        """

        :param user_name:
        :return:
        """

        delete_user = self.get_user_by_name(user_name)

        logger.info(f"Deleting {user_name}")
        with self.plan.target.connect():
            self.plan.target.server.users.remove(delete_user.id)

        logger.success("Deleted user")
        return True

    def update_user(self, user_name: str, role: str) -> bool:
        """

        :param user_name:
        :param role:
        :return:
        """
        logger.info(f"Updating user: {user_name}")
        update_user = self.get_user_by_name(user_name)

        update_user.site_role = role

        with self.plan.target.connect():
            self.plan.target.server.users.update(update_user)

        logger.success("User updated")
        return True

    def get_user_by_name(self, user_name: str) -> TSC.UserItem:
        """

        :param user_name:
        :return:
        """
        # Sets up the request param to filter for the correct username.
        req_option = TSC.RequestOptions()
        req_option.filter.add(
            TSC.Filter(
                TSC.RequestOptions.Field.Name,
                TSC.RequestOptions.Operator.Equals,
                user_name,
            )
        )
        with self.plan.target.connect():
            all_users, pagination_item = self.plan.target.server.users.get(
                req_options=req_option
            )

            if pagination_item.total_available > 1:
                raise ValueError(
                    f"Multiple user names have been returned, {pagination_item.total_available}."
                )

            if pagination_item.total_available < 1:
                logger.warning(
                    f"No user was found matching the user name {user_name} on the site."
                )
                return

            return all_users[0]

    def compare_dicts(self, reference: Dict, target: Dict) -> [List, List, Dict]:
        """

        :param reference:
        :param target:
        :return:
        """
        keys_only_in_reference = [key for key in reference if key not in target]
        keys_only_in_target = [key for key in target if key not in reference]
        keys_with_different_values = [
            (key, reference[key])
            for key in reference
            if key in target and reference[key] != target[key]
        ]

        return keys_only_in_reference, keys_only_in_target, keys_with_different_values



"""Module for classes and functions for managing Tableau Workbooks"""
import os
from typing import Optional
from TableauConMan.assets_manager import AssetsManager
from TableauConMan.plan import Plan
from TableauConMan.data_sources_manager import DataSourcesManager
import tableauserverclient as TSC
from loguru import logger


class WorkbooksManager(AssetsManager):
    """
    WorkbooksManager

    This class is responsible for managing workbooks in a Tableau Server environment.

    Attributes:
        - asset_type: The type of assets managed by this manager (Workbooks)
        - target_workbooks: A list of target workbook items fetched from the target server
        - target_workbooks_list: A list of target workbook names
        - reference_workbooks: A list of reference workbook items fetched from the reference server
        - reference_workbooks_list: A list of reference workbook names

    Methods:
        - populate_workbooks(): Fetches target and reference workbooks and generates their respective lists
        - _generate_server_list(source, asset_type, request_filter=None): Generates a list of server assets for the given source and asset type
        - get_workbook_options(): Retrieves workbook options from the asset options dictionary in the plan
        - get_workbook_changes(): Compares target and reference workbooks, returning lists of common, to remove, and to add workbooks
        - delete(server, workbook): Deletes the given workbook from the server
        - download(server, workbook): Downloads the given workbook from the server and returns the file path
        - format_target_workbook(reference_workbook): Formats a reference workbook item to match the structure of a target workbook
        - get_target_project(): Retrieves the target project ID based on the plan asset project_type option
        - upload(server, workbook, file_path): Uploads the given workbook to the server and returns the uploaded workbook
        - add(list_to_add, list_to_update=[], workbook_options={}): Adds and updates workbooks from the given lists by downloading reference workbooks,
          setting the project and uploading them to the target server
        - remove(remove_list): Removes workbooks from the target server based on the given list of workbook names

    Note:
        - This class extends the AssetsManager class
        - The methods in this class require a valid Tableau Server connection to function properly
        - Handling of exceptions and logging is not included in this class
    """

    def __init__(self, plan) -> None:
        """

        :param plan:
        """
        AssetsManager.__init__(self, plan)
        self.asset_type = self.AssetType.Workbooks
        self.target_workbooks: Optional[list[TSC.WorkbookItem]] = None
        self.target_workbooks_list: Optional[list] = None
        self.reference_workbooks: Optional[list[TSC.WorkbookItem]] = None
        self.reference_workbooks_list: Optional[list] = None

    def populate_workbooks(self):
        """
        Use object properties to get a comparable list of workbooks from the source and reference
        """
        self.target_workbooks, self.target_workbooks_list = self._generate_server_list(
            self.plan.target, "workbooks", self.plan.target_selection_rules
        )

        (
            self.reference_workbooks,
            self.reference_workbooks_list,
        ) = self._generate_server_list(
            self.plan.reference, "workbooks", self.plan.reference_selection_rules
        )

    def _generate_server_list(
        self, source, asset_type, request_filter: Optional[TSC.RequestOptions] = None
    ):
        """

        :param source:
        :param asset_type:
        :param filter:
        :return:
        """
        asset = getattr(source.server, asset_type)

        with source.connect():
            asset_items = list(TSC.Pager(asset, request_filter))
            # print(workbooks)
            """
          Generalized: can be moved to parent class
          Find a way to included updated_at check.  Might be able to use dict to
          """

            asset_list = []
            for asset in asset_items:
                # print(workbook.name)
                asset_list.append(asset.name)
            # print(workbook_list)
            return asset_items, asset_list

    def get_workbook_options(self):
        """

        :return:
        """
        workbook_options = self.plan.asset_options.get("workbooks")

        return workbook_options

    def get_workbook_changes(self):
        """

        :return:
        """
        common, to_remove, to_add = self.get_changes(
            self.target_workbooks_list, self.reference_workbooks_list
        )

        update = []
        for workbook_name in common:
            target_workbook = self.get_item_from_list(
                workbook_name, self.target_workbooks
            )

            reference_workbook = self.get_item_from_list(
                workbook_name, self.reference_workbooks
            )

            # logger.debug(f"Reference Updated At: {reference_datasource.updated_at.strftime('%Y-%m-%d, %H:%M:%S')} | Target Updated At {target_datasource.updated_at.strftime('%Y-%m-%d, %H:%M:%S')}")

            if reference_workbook.updated_at > target_workbook.updated_at:
                update.append(workbook_name)

        return update, to_add, to_remove

    def delete(self, server, workbook):
        """

        :param server:
        :param workbook:
        """
        with server.connect():
            server.server.workbooks.delete(workbook.id)

    def download(self, server, workbook):
        """

        :param server:
        :param workbook:
        :return:
        """
        with server.connect():
            file_path = server.server.workbooks.download(workbook.id)

        return file_path

    def format_target_workbook(self, reference_workbook):
        """

        :param reference_workbook:
        :return:
        """
        target_workbook = TSC.WorkbookItem("")
        target_workbook.name = reference_workbook.name
        target_workbook.project_id = self.get_target_project()

        return target_workbook

    def get_target_project(self):
        """
        Get the target project based on the plan asset project_type option
        """
        workbook_options = self.get_workbook_options()
        workbook_mapping = workbook_options.get("mapping")
        request_options = self.plan.format_rule(workbook_mapping.get("project_filter"))
        server = self.plan.target
        with server.connect():
            server_project = self.plan.target.server.projects.get(request_options)[0][0]
        target_project = server_project.id
        return target_project

    def upload(self, server, workbook, file_path):
        """

        :param server:
        :param workbook:
        :param file_path:
        :return:
        """
        with server.connect():
            new_workbook = server.server.workbooks.publish(
                workbook,
                file_path,
                "Overwrite",
                skip_connection_check=True,
            )

        return new_workbook

    def add(
        self,
        list_to_add: list,
        list_to_update: list = [],
        workbook_options: dict = {},
    ):
        """
        Match the workbook from list to reference object
        Download reference workbook
        Get source project object
        Set workbook project to source project_id
        Upload workbook
        Delete downloaded workbook file
        """
        to_add_and_update = list_to_add + list_to_update
        processed_wordbooks_count = 0

        if len(to_add_and_update) > 0:
            for workbook_name in to_add_and_update:
                reference_workbook = self.get_item_from_list(
                    workbook_name, self.reference_workbooks
                )

                # populate connections and look at what if it matches the known datasources

                file_path = self.download(self.plan.reference, reference_workbook)
                target_workbook = self.format_target_workbook(reference_workbook)
                try:
                    uploaded_workbook = self.upload(
                        self.plan.target, target_workbook, file_path
                    )

                    self.update_connection_credentials(
                        uploaded_workbook, self.AssetType.Workbooks
                    )
                    logger.info(
                        f"Processed Datasource Connections for: {workbook_name}"
                    )
                    processed_wordbooks_count += 1
                except Exception as exc:
                    logger.warning(
                        f"There was a problem processing workbook {workbook_name}: {exc}"
                    )
                finally:
                    os.remove(file_path)
                    continue

        logger.info(f"Added {processed_wordbooks_count} Workbooks")

    def remove(self, remove_list: list):
        """
        Match the workbook from list to reference object
        Delete Workbook
        """
        if len(remove_list) > 0:
            for workbook_name in remove_list:
                workbook = self.get_item_from_list(workbook_name, self.target_workbooks)

                self.delete(self.plan.target, workbook)
                logger.info(f"Workbook {workbook.name} deleted")
        logger.info(f"Removed {len(remove_list)} Workbooks")



"""Module for managing provided specification files"""
from typing import Dict, List
import cerberus
import yaml

from TableauConMan.schemas import spec as sch

VALIDATION_ERR_MSG = 'Spec error: {} "{}", field "{}": {}'


class IndentDumper(yaml.Dumper):
    """
    The `IndentDumper` class is a subclass of `yaml.Dumper` that provides a customized implementation of the `increase_indent` method.

    This class defines a method `increase_indent` which overrides the method of the same name in the parent class `yaml.Dumper`. The `increase_indent` method adjusts the indentation level
    * used when serializing YAML data.

    Parameters:
        - `flow` (bool): Whether the current context is a flow context (default: False).
        - `indentless` (bool): Whether the current context does not require indentation (default: False).

    Returns:
        - `int`: The new indentation level.

    Usage:
        To use the `IndentDumper` class, you can create an instance of it and pass it as an argument when serializing YAML data using the `yaml.dump` function.

    Example:
        import yaml

        # Create an instance of IndentDumper
        dumper = IndentDumper()

        # Serialize YAML data using the custom dumper
        yaml.dump(data, Dumper=dumper)
    """

    def increase_indent(self, flow=False, indentless=False):
        return super(IndentDumper, self).increase_indent(flow, False)


class Specification:

    """
    Load the specification from a raw dictionary.

    :param raw_spec: A dictionary containing the specification.
    :raises TypeError: If the raw_spec is not a dictionary.
    :raises ValueError: If the raw_spec does not have a valid schema.
    """

    def __init__(self):
        self.site: dict = None
        self.projects: list = None
        self.permission_templates: list = None
        self.groups: list = None
        self.users: list = None
        self.raw_spec: dict = None

    def load_spec(self, raw_spec: dict) -> None:
        """

        :param raw_plan:
        """
        if not isinstance(raw_spec, dict):
            raise TypeError("The raw spec must be a dictionary")

        error_messages = self.ensure_valid_schema(raw_spec)
        if error_messages:
            raise ValueError("\n".join(error_messages))

        self.site = raw_spec.get("site")
        self.projects = raw_spec.get("projects")
        self.permission_templates = raw_spec.get("permission_templates")
        self.groups = raw_spec.get("groups")
        self.users = raw_spec.get("users")
        self.raw_spec = raw_spec

    @staticmethod
    def ensure_valid_schema(spec: Dict) -> List[str]:
        """
        Ensure that the provided spec has no schema errors.

        Returns a list with all the errors found.
        """
        error_messages = []

        validator = cerberus.Validator(yaml.safe_load(sch.SPEC_SCHEMA))
        validator.validate(spec)
        for entity_type, err_msg in validator.errors.items():
            if isinstance(err_msg[0], str):
                error_messages.append(f"Spec error: {entity_type}: {err_msg[0]}")
                continue

            for error in err_msg[0].values():
                error_messages.append(f"Spec error: {entity_type}: {error[0]}")
        if error_messages:
            return error_messages

        schema = {
            "projects": yaml.safe_load(sch.SPEC_PROJECTS_SCHEMA),
            "permission_templates": yaml.safe_load(
                sch.SPEC_PERMISSION_TEMPLATES_SCHEMA
            ),
            "groups": yaml.safe_load(sch.SPEC_GROUPS_SCHEMA),
            "users": yaml.safe_load(sch.SPEC_USERS_SCHEMA),
        }

        validators = {
            "projects": cerberus.Validator(schema.get("projects")),
            "permission_templates": cerberus.Validator(
                schema.get("permission_templates")
            ),
            "groups": cerberus.Validator(schema.get("groups")),
            "users": cerberus.Validator(schema.get("users")),
        }

        entities_by_type = []
        for entity_type, entities in spec.items():
            if entities and entity_type in [
                "projects",
                "permission_templates",
                "groups",
                "users",
            ]:
                entities_by_type.append((entity_type, entities))

        for entity_type, entities in entities_by_type:
            for entity_dict in entities:
                validators[entity_type].validate(entity_dict)
                for field, err_msg in validators[entity_type].errors.items():
                    error_messages.append(
                        VALIDATION_ERR_MSG.format(
                            entity_type, entity_dict.get("name"), field, err_msg[0]
                        )
                    )

        return error_messages

    def write_spec(self, dict_file, filename):
        with open(f"{filename}.yaml", "w") as file:
            documents = yaml.dump(
                dict_file,
                file,
                default_flow_style=False,
                sort_keys=False,
                Dumper=IndentDumper,
            )



FIND_WORKBOOK_DATASOURCES = """
{
  "data": {
    "workbooksConnection": {
      "nodes": [
        {
          "name": " ",
          "upstreamDatasources": [
            {
              "name": " ",
              "luid": " "
            }
          ]
        },
        {
          "name": " ",
          "upstreamDatasources": [
            {
              "name": " ",
              "luid": " "
            },
            
          ]
        }
      ]
    }
  }
}
"""



"""A module to connect to yaml files or yaml strings"""
import os
from typing import Optional
import yaml
from TableauConMan.helpers import utils


class YamlConnector:
    """
    YamlConnector Class

    Handles reading and parsing of YAML files or strings.

    Attributes:
        file_path (Optional[str]): The file path of the YAML file.
        yaml_string (Optional[str]): The YAML string.

    Methods:
        get_yaml(): Returns the parsed YAML data.

    """

    def __init__(
        self, file_path: Optional[str] = None, yaml_string: Optional[str] = None
    ):
        """

        :param file_path:
        :param yaml:
        """
        self._file_path = file_path
        self._yaml_string = yaml_string

    @property
    def file_path(self) -> Optional[str]:
        """

        :return:
        """
        return self._file_path

    @file_path.setter
    def file_path(self, value):
        """

        :param value:
        """
        self._file_path = value

    @property
    def yaml_string(self) -> Optional[str]:
        """

        :return:
        """
        return self._yaml_string

    @yaml_string.setter
    def yaml_string(self, value):
        """

        :param value:
        """
        self._yaml_string = value

    def get_yaml(self):
        """

        :param object:
        :return:
        """
        if self._file_path is not None:
            file_path_parts = os.path.split(self._file_path)
            search_path = file_path_parts[0]
            file_name = file_path_parts[1]
            jinja_environment = utils.get_jinja_environment(search_path)

            try:
                file_text_list = yaml.safe_load(
                    jinja_environment.get_template(file_name).render()
                )
                return file_text_list
            except yaml.YAMLError as exc:
                print(exc)
                return None

        elif self._yaml_string is not None:
            jinja_environment = utils.get_jinja_environment("string")
            file_text = self._yaml_string
            try:
                file_text_list = yaml.safe_load(
                    jinja_environment.from_string(file_text).render()
                )
                return file_text_list
            except yaml.YAMLError as exc:
                print(exc)
                return None
